# Copyright Okera Inc. All Rights Reserved.
#
# Tests for KV store API
#

import copy
import os
import unittest

from okera.tests import pycerebro_test_common as common
from okera._thrift_api import TKvStoreGetParams, TKvStorePutParams, TRecordServiceException

class KvStoreTest(common.TestBase):
    def _get_kv(self, conn, key):
        request = TKvStoreGetParams()
        request.key = key
        return conn.service.client.GetFromKvStore(request).value

    def _put_kv(self, conn, key, value):
        request = TKvStorePutParams()
        request.key = key
        request.value = value
        return conn.service.client.PutToKvStore(request)

    def test_basic(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._put_kv(conn, 'k1', 'value')
            value = self._get_kv(conn, 'k1')
            self.assertEqual(value, 'value')

            self._put_kv(conn, 'k1', 'value2')
            value = self._get_kv(conn, 'k1')
            self.assertEqual(value, 'value2')

            self._put_kv(conn, 'k1', '')
            value = self._get_kv(conn, 'k1')
            self.assertEqual(value, '')

            value = self._get_kv(conn, 'k2')
            self.assertEqual(value, None)

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for role granting
#
# pylint: disable=too-many-public-methods
# pylint: disable=line-too-long
# pylint: disable=len-as-condition
# pylint: disable=too-many-locals

import unittest

from okera._thrift_api import TAuthorizeQueryParams
from okera.tests import pycerebro_test_common as common

DB = 'test_privacy_db'
ROLE = 'test_privacy_role'
WORKSPACE_ROLE = 'okera_workspace_role'
VIEW = 'test_privacy_view'
USER1 = 'testprivacyuser1'
USER2 = 'testprivacyuser2'
USER3 = 'testprivacyuser3'

ATTR1 = "test_privacy_attr.attr1"
ATTR2 = "test_privacy_attr.attr2"

ALLTYPES_TABLE = 'alltypes'
ALLTYPES_TABLE2 = 'alltypes2'
ALLTYPES_ALL_ROWS_VIEW = 'alltypes_all_rows_view'
ALLTYPES_FIRST_ROW_DOUBLED_VIEW = 'alltypes_first_row_doubled_view'
ALLTYPES_SECOND_ROW_DOUBLED_VIEW = 'alltypes_second_row_doubled_view'
ALLTYPES_PRIVACY_VIEW = 'alltypes_privacy_view'
ALLTYPES_PRIVACY_UNIQUE_VIEW = 'alltypes_privacy_unique_view'
ALLTYPES_PRIVACY_SHARED_VIEW = 'alltypes_privacy_shared_view'
ALLTYPES_PRIVACY_UNIQUE_VIEW1 = 'alltypes_privacy_unique_no_ref_int_view1'
ALLTYPES_PRIVACY_UNIQUE_VIEW2 = 'alltypes_privacy_unique_no_ref_int_view2'
ALLTYPES_PRIVACY_SHARED_VIEW1 = 'alltypes_privacy_shared_no_ref_int_view1'
ALLTYPES_PRIVACY_SHARED_VIEW2 = 'alltypes_privacy_shared_no_ref_int_view2'

ALLTYPES_S3_DDL = """
CREATE EXTERNAL TABLE %s.%%s (
  bool_col BOOLEAN ATTRIBUTE test_privacy_attr.attr1 test_privacy_attr.attr2,
  tinyint_col TINYINT ATTRIBUTE test_privacy_attr.attr1 test_privacy_attr.attr2,
  smallint_col SMALLINT ATTRIBUTE test_privacy_attr.attr1 test_privacy_attr.attr2,
  int_col INT ATTRIBUTE test_privacy_attr.attr1 test_privacy_attr.attr2,
  bigint_col BIGINT ATTRIBUTE test_privacy_attr.attr1 test_privacy_attr.attr2,
  float_col FLOAT ATTRIBUTE test_privacy_attr.attr1,
  double_col DOUBLE ATTRIBUTE test_privacy_attr.attr1,
  string_col STRING ATTRIBUTE test_privacy_attr.attr1,
  varchar_col VARCHAR(10) ATTRIBUTE test_privacy_attr.attr1,
  char_col CHAR(5) ATTRIBUTE test_privacy_attr.attr1,
  timestamp_col TIMESTAMP ATTRIBUTE test_privacy_attr.attr1,
  decimal_col DECIMAL(24,10) ATTRIBUTE test_privacy_attr.attr1
)
ROW FORMAT DELIMITED FIELDS TERMINATED BY '|'
WITH SERDEPROPERTIES ('field.delim'='|', 'serialization.format'='|')
STORED AS TEXTFILE
LOCATION 's3a://cerebrodata-test/alltypes'
""" % (DB)

ALLTYPES_ALL_ROWS_VIEW_DDL = """
CREATE VIEW %s.%s AS SELECT * FROM %s.%s
""" % (DB, ALLTYPES_ALL_ROWS_VIEW, DB, ALLTYPES_TABLE)

ALLTYPES_PRIVACY_QUERY = """
SELECT
    {privacy_fn}(bool_col) as bool_col,
    {privacy_fn}(tinyint_col) as tinyint_col,
    {privacy_fn}(smallint_col) as smallint_col,
    {privacy_fn}(int_col) as int_col,
    {privacy_fn}(bigint_col) as bigint_col,
    {privacy_fn}(float_col) as float_col,
    {privacy_fn}(double_col) as double_col,
    {privacy_fn}(string_col) as string_col,
    {privacy_fn}(varchar_col) as varchar_col,
    {privacy_fn}(char_col) as char_col,
    {privacy_fn}(timestamp_col) as timestamp_col,
    {privacy_fn}(decimal_col) as decimal_col
FROM {db}.{table}
"""

ALLTYPES_PRIVACY_VIEW_DDL = """
CREATE VIEW {db}.{view} AS SELECT
    {privacy_fn}(bool_col) as bool_col,
    {privacy_fn}(tinyint_col) as tinyint_col,
    {privacy_fn}(smallint_col) as smallint_col,
    {privacy_fn}(int_col) as int_col,
    {privacy_fn}(bigint_col) as bigint_col,
    {privacy_fn}(float_col) as float_col,
    {privacy_fn}(double_col) as double_col,
    {privacy_fn}(string_col) as string_col,
    {privacy_fn}(varchar_col) as varchar_col,
    {privacy_fn}(char_col) as char_col,
    {privacy_fn}(timestamp_col) as timestamp_col,
    {privacy_fn}(decimal_col) as decimal_col
FROM {db}.{table}
"""

ALLTYPES_FIRST_ROW_DOUBLED_VIEW_DDL = """
CREATE VIEW %s.%s AS %s
WHERE bool_col = true
UNION ALL %s
WHERE bool_col = true
""" % (DB, ALLTYPES_FIRST_ROW_DOUBLED_VIEW, ALLTYPES_PRIVACY_QUERY, ALLTYPES_PRIVACY_QUERY)

ALLTYPES_SECOND_ROW_DOUBLED_VIEW_DDL = """
CREATE VIEW %s.%s AS %s
WHERE bool_col = false
UNION ALL %s
WHERE bool_col = false
""" % (DB, ALLTYPES_SECOND_ROW_DOUBLED_VIEW, ALLTYPES_PRIVACY_QUERY, ALLTYPES_PRIVACY_QUERY)

ALLTYPES_PRIVACY_UNIQUE_VIEW_DDL = """
CREATE VIEW {db}.{view} AS SELECT
    {privacy_fn}(bool_col, signed_user(true)) as bool_col,
    {privacy_fn}(tinyint_col, signed_user(true)) as tinyint_col,
    {privacy_fn}(smallint_col, signed_user(true)) as smallint_col,
    {privacy_fn}(int_col, signed_user(true)) as int_col,
    {privacy_fn}(bigint_col, signed_user(true)) as bigint_col,
    {privacy_fn}(float_col, signed_user(true)) as float_col,
    {privacy_fn}(double_col, signed_user(true)) as double_col,
    {privacy_fn}(string_col, signed_user(true)) as string_col,
    {privacy_fn}(varchar_col, signed_user(true)) as varchar_col,
    {privacy_fn}(char_col, signed_user(true)) as char_col,
    {privacy_fn}(timestamp_col, signed_user(true)) as timestamp_col,
    {privacy_fn}(decimal_col, signed_user(true)) as decimal_col
FROM {db}.{table}
"""

ALLTYPES_PRIVACY_SHARED_VIEW_DDL = """
CREATE VIEW {db}.{view} AS SELECT
    {privacy_fn}(bool_col, signed_user(false)) as bool_col,
    {privacy_fn}(tinyint_col, signed_user(false)) as tinyint_col,
    {privacy_fn}(smallint_col, signed_user(false)) as smallint_col,
    {privacy_fn}(int_col, signed_user(false)) as int_col,
    {privacy_fn}(bigint_col, signed_user(false)) as bigint_col,
    {privacy_fn}(float_col, signed_user(false)) as float_col,
    {privacy_fn}(double_col, signed_user(false)) as double_col,
    {privacy_fn}(string_col, signed_user(false)) as string_col,
    {privacy_fn}(varchar_col, signed_user(false)) as varchar_col,
    {privacy_fn}(char_col, signed_user(false)) as char_col,
    {privacy_fn}(timestamp_col, signed_user(false)) as timestamp_col,
    {privacy_fn}(decimal_col, signed_user(false)) as decimal_col
FROM {db}.{table}
"""

GRANT_WITH_TRANSFORM_ATTR1_DDL = """
GRANT SELECT ON DATABASE %s
TRANSFORM %s WITH %%s
TO ROLE %s
""" % (DB, ATTR1, ROLE)

REVOKE_WITH_TRANSFORM_ATTR2_DDL = """
REVOKE SELECT ON TABLE %s.%s
TRANSFORM %s WITH %%s
FROM ROLE %s
""" % (DB, ALLTYPES_TABLE, ATTR1, ROLE)

GRANT_WITH_TRANSFORM_ATTR2_DDL = """
GRANT SELECT ON DATABASE %s
TRANSFORM %s WITH %%s
TO ROLE %s
""" % (DB, ATTR2, ROLE)

REVOKE_WITH_TRANSFORM_ATTR2_DDL = """
REVOKE SELECT ON TABLE %s.%s
TRANSFORM %s WITH %%s
FROM ROLE %s
""" % (DB, ALLTYPES_TABLE, ATTR2, ROLE)

ALLTYPES_STRING_COLS = ['varchar_col', 'string_col', 'char_col', 'timestamp_col']
ALLTYPES_TRANSFORMED_COLS = ['bool_col', 'tinyint_col', 'smallint_col', 'int_col', 'bigint_col']
ALLTYPES_RESULTS = [
    {'float_col': 4.0, 'varchar_col': 'vchar1', 'decimal_col': '3.1415920000',
     'smallint_col': 1, 'double_col': 5.0, 'bool_col': True, 'string_col': 'hello',
     'int_col': 2, 'timestamp_col': '2015-01-01 00:00:00.000', 'bigint_col': 3,
     'char_col': 'char1', 'tinyint_col': 0},
    {'float_col': 10.0, 'varchar_col': 'vchar2', 'decimal_col': '1234.5678900000',
     'smallint_col': 7, 'double_col': 11.0, 'bool_col': False, 'string_col': 'world',
     'int_col': 8, 'timestamp_col': '2016-01-01 00:00:00.000', 'bigint_col': 9,
     'char_col': 'char2', 'tinyint_col': 6}
]

ALLTYPES_CONSTANT_VALUES = {
    'bool_col': 'true',
    'tinyint_col': 'cast(5 AS TINYINT)',
    'smallint_col': 'cast(1024 AS TINYINT)',
    'int_col': 'cast(12345678 AS TINYINT)',
    'bigint_col': 'cast(123456789123456 AS BIGINT)',
    'float_col': 'cast(12.345 AS FLOAT)',
    'double_col': 'cast(12.345 AS DOUBLE)',
    'string_col': '\'abcdefghi\'',
    'varchar_col': 'cast(\'abcde\' AS VARCHAR(10))',
    'char_col': 'cast(\'abcde\' AS CHAR(5))',
    'timestamp_col': 'cast(10 AS TIMESTAMP)',
    'decimal_col': 'cast(12.345 AS DECIMAL(5,3))',
}

def scan_as_json(ctx, conn, query, dialect='okera', user=None):
    current_user = None
    if user:
        current_user = ctx.get_token()
        ctx.enable_token_auth(token_str=user)
    res = conn.scan_as_json(query, dialect=dialect)

    if current_user:
        ctx.enable_token_auth(token_str=current_user)
    else:
        ctx.disable_auth()
    return res

def authorize_query(conn, query, user):
    request = TAuthorizeQueryParams()
    request.sql = query
    request.requesting_user = user
    result = conn.service.client.AuthorizeQuery(request)
    return ' '.join(result.result_sql.split())

class PrivacyTest(unittest.TestCase):
    def _validate_query(self, ctx, conn, query,
                        equal_across_users=True, equal_across_queries=True,
                        has_results=True, equal_across_rows=False, null_col=None,
                        check_results=True):
        print(query)
        user1_res1 = scan_as_json(ctx, conn, query, user=USER1)
        user1_res2 = scan_as_json(ctx, conn, query, user=USER1)
        user2_res1 = scan_as_json(ctx, conn, query, user=USER2)

        if not check_results:
            return

        if has_results:
            self.assertTrue(len(user1_res1) > 0)
            self.assertTrue(len(user1_res2) > 0)
            self.assertTrue(len(user2_res1) > 0)
            if null_col:
                for result in [user1_res1, user1_res2, user2_res1]:
                    for record in result:
                        self.assertTrue(record[null_col] is None)
        else:
            self.assertTrue(len(user1_res1) == 0)
            self.assertTrue(len(user1_res2) == 0)
            self.assertTrue(len(user2_res1) == 0)

        if has_results:
            # These checks are only meaningful if we have results
            if equal_across_queries:
                self.assertTrue(user1_res1 == user1_res2)
            else:
                self.assertTrue(user1_res1 != user1_res2)

            if equal_across_users:
                self.assertTrue(user1_res1 == user2_res1)
            else:
                self.assertTrue(user1_res1 != user2_res1)

            if equal_across_rows:
                self.assertTrue(len(user1_res1) == 2)
                self.assertTrue(len(user1_res2) == 2)
                self.assertTrue(len(user2_res1) == 2)
                self.assertTrue(user1_res1[0] == user1_res1[1])
                self.assertTrue(user1_res2[0] == user1_res2[1])
                self.assertTrue(user2_res1[0] == user2_res1[1])


    def _test_fn(self, fn, maintains_ref_integrity, stable, maintains_type):
        PRIVACY_FN = fn
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP ATTRIBUTE IF EXISTS %s' % ATTR1)
            conn.execute_ddl('DROP ATTRIBUTE IF EXISTS %s' % ATTR2)
            conn.execute_ddl('DROP ROLE IF EXISTS %s' % ROLE)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
            conn.execute_ddl('CREATE DATABASE %s' % DB)
            conn.execute_ddl('CREATE ATTRIBUTE %s' % ATTR1)
            conn.execute_ddl('CREATE ATTRIBUTE %s' % ATTR2)

            conn.execute_ddl(ALLTYPES_S3_DDL % ALLTYPES_TABLE)
            conn.execute_ddl(ALLTYPES_S3_DDL % ALLTYPES_TABLE2)
            conn.execute_ddl(ALLTYPES_FIRST_ROW_DOUBLED_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_SECOND_ROW_DOUBLED_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, view=ALLTYPES_PRIVACY_VIEW,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_UNIQUE_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, view=ALLTYPES_PRIVACY_UNIQUE_VIEW,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_SHARED_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, view=ALLTYPES_PRIVACY_SHARED_VIEW,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_SHARED_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, view=ALLTYPES_PRIVACY_SHARED_VIEW1,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_SHARED_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE2, view=ALLTYPES_PRIVACY_SHARED_VIEW2,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_UNIQUE_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE, view=ALLTYPES_PRIVACY_UNIQUE_VIEW1,
                privacy_fn=PRIVACY_FN))
            conn.execute_ddl(ALLTYPES_PRIVACY_UNIQUE_VIEW_DDL.format(
                db=DB, table=ALLTYPES_TABLE2, view=ALLTYPES_PRIVACY_UNIQUE_VIEW2,
                privacy_fn=PRIVACY_FN))

            conn.execute_ddl('CREATE ROLE %s' % ROLE)
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (ROLE, USER1))
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (ROLE, USER2))
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (ROLE, USER3))
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (WORKSPACE_ROLE, USER1))
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (WORKSPACE_ROLE, USER2))
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (WORKSPACE_ROLE, USER3))

            conn.execute_ddl(GRANT_WITH_TRANSFORM_ATTR2_DDL % (PRIVACY_FN + '()'))

            res = scan_as_json(ctx, conn,
                               'SELECT * FROM %s.%s' % (DB, ALLTYPES_TABLE))
            self.assertTrue(res == ALLTYPES_RESULTS)

            # Verify that for authorize query, no seed is added to the privacy
            # function.
            authorized_sql = authorize_query(
                conn, 'SELECT * FROM %s.%s' % (DB, ALLTYPES_TABLE), USER1)
            expected = ("SELECT %s(bool_col) as bool_col, " +
                        "%s(tinyint_col) as tinyint_col, " +
                        "%s(smallint_col) as smallint_col, " +
                        "%s(int_col) as int_col, " +
                        "%s(bigint_col) as bigint_col, " +
                        "float_col, double_col, string_col, varchar_col, " +
                        "char_col, timestamp_col, decimal_col " +
                        "FROM test_privacy_db.alltypes") %\
                         (fn, fn, fn, fn, fn)
            self.assertEqual(authorized_sql, expected)

            # When executed against a table with a grant, should see the same thing
            # by default
            self._validate_query(ctx, conn,
                                 'SELECT * FROM %s.%s' % (DB, ALLTYPES_TABLE),
                                 equal_across_queries=stable, equal_across_users=stable,
                                 has_results=True)
            join_query = '''SELECT a.*, b.* FROM {db}.{table} a
                 JOIN {db}.{table} b ON a.bigint_col=b.bigint_col'''.format(
                     db=DB, table=ALLTYPES_TABLE)
            self._validate_query(ctx, conn, join_query, equal_across_queries=stable,
                                 equal_across_users=stable, has_results=stable)

            # Outer join should always return the outer table row even when there are
            # no matches but the other side should be null.
            join_query = '''SELECT a.*, b.bigint_col as be_null FROM {db}.{table} a
                 LEFT OUTER JOIN {db}.{table} b ON a.bigint_col=b.bigint_col'''.format(
                     db=DB, table=ALLTYPES_TABLE)
            null_col = 'be_null'
            if stable:
                null_col = None
            self._validate_query(ctx, conn, join_query, equal_across_queries=stable,
                                 equal_across_users=stable, has_results=True,
                                 null_col=null_col)

            # When executed against a view, should see the same thing by default
            self._validate_query(ctx, conn,
                                 'SELECT * FROM %s.%s' % (DB, ALLTYPES_PRIVACY_VIEW),
                                 equal_across_queries=stable, equal_across_users=stable,
                                 has_results=True)
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_VIEW)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=stable, equal_across_users=stable,
                                 has_results=stable)

            # When executed against a view with duplicate rows, each row should either
            # contain the same values if the function is stable, or not if it is not stable
            self._validate_query(ctx, conn,
                                 'SELECT * FROM %s.%s' % (DB, ALLTYPES_FIRST_ROW_DOUBLED_VIEW),
                                 equal_across_queries=stable, equal_across_users=stable, equal_across_rows=stable, has_results=True)

            # When executed against a view that is shared, should see the same thing by
            # default
            self._validate_query(ctx, conn,
                                 'SELECT * FROM %s.%s' % (DB, ALLTYPES_PRIVACY_SHARED_VIEW),
                                 equal_across_queries=stable, equal_across_users=stable,
                                 has_results=True)
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_SHARED_VIEW)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=stable, equal_across_users=stable,
                                 has_results=stable)

            # When executed against a view that is unique, should see the same thing
            # across runs but not across users
            self._validate_query(ctx, conn,
                                 'SELECT * FROM %s.%s' % (DB, ALLTYPES_PRIVACY_UNIQUE_VIEW),
                                 equal_across_queries=stable, equal_across_users=False, has_results=True)
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_UNIQUE_VIEW)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=stable, equal_across_users=False, has_results=stable)

            # Check referential integrity, which is dependent on which security function we use.
            # Note that the view is unique per user, so will never have the same results across users
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table2} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_UNIQUE_VIEW1, table2=ALLTYPES_PRIVACY_UNIQUE_VIEW2)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=stable, equal_across_users=False, has_results=maintains_ref_integrity)

            # Check referential integrity, which is dependent on which security function we use.
            # Note that the view is shared for all users, so will have the same results across users
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table2} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_SHARED_VIEW1, table2=ALLTYPES_PRIVACY_SHARED_VIEW2)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=stable, equal_across_users=maintains_ref_integrity, has_results=maintains_ref_integrity)

            # Check referential integrity, which is dependent on which security function we use.
            # This checks it across different parameters to the same privacy function, which means
            # there is no match.
            join_query = 'SELECT a.*, b.* FROM {db}.{table} a JOIN {db}.{table2} b ON a.bigint_col=b.bigint_col'.format(
                db=DB, table=ALLTYPES_PRIVACY_SHARED_VIEW1, table2=ALLTYPES_PRIVACY_UNIQUE_VIEW2)
            self._validate_query(ctx, conn, join_query,
                                 equal_across_queries=False, equal_across_users=False, has_results=False)

            # Test constant conversions - there's too much randomness here to check the actual
            # values consistently across all functions, but we can at least check to make sure
            # we don't error out.
            cols = []
            for col, expr in ALLTYPES_CONSTANT_VALUES.items():
                cols.append('%s(%s) as %s' % (fn, expr, col))
            query = 'SELECT %s' % (','.join(cols))
            self._validate_query(ctx, conn, query, check_results=False)

            for key, val in ALLTYPES_RESULTS[0].items():
                # These columns are too low cardinality to reliably pass
                if key in ['bool_col', 'tinyint_col']:
                    continue

                # We only transform some values in ALLTYPES_TABLE (via the transform policy),
                # and if we don't transform them, we do expect values to come back
                not_transformed = key not in ALLTYPES_TRANSFORMED_COLS
                if key in ALLTYPES_STRING_COLS:
                    val = '\'%s\'' % val
                # Check that the filter has tokenization applied on it properly
                join_query = 'SELECT * FROM {db}.{table} WHERE {key}={value}'.format(
                    db=DB, table=ALLTYPES_TABLE, key=key, value=val)
                self._validate_query(ctx, conn, join_query,
                                     equal_across_queries=stable, equal_across_users=stable, has_results=not_transformed)

                # Check that the aggregation has tokenization applied on it properly
                # Note that we explicitly reference one of the transform keys to ensure the result
                # set is appropriately stable/not stable
                join_query = 'SELECT {key}, {transformed_key}, max({key}) FROM {db}.{table} GROUP BY 1, 2 HAVING max({key})={value}'.format(
                    db=DB, table=ALLTYPES_TABLE, key=key, value=val, transformed_key='bigint_col')
                self._validate_query(ctx, conn, join_query,
                                     equal_across_queries=stable, equal_across_users=stable, has_results=not_transformed)

                # Check that the group by having has tokenization applied on it properly
                # Note that we explicitly reference one of the transform keys to ensure the result
                # set is appropriately stable/not stable
                join_query = 'SELECT {key}, {transformed_key}, count(*) FROM {db}.{table} GROUP BY 1, 2 HAVING {key}={value}'.format(
                    db=DB, table=ALLTYPES_TABLE, key=key, value=val, transformed_key='bigint_col')
                self._validate_query(ctx, conn, join_query,
                                     equal_across_queries=stable, equal_across_users=stable, has_results=not_transformed)

                # We can only do these checks if we have a format-preserving privacy function,
                # otherwise the comparison fails the type check in analysis, or we can do this
                # if the types are equivalent-ish, which we have encoded as whether we've transformed
                # this column in the transform grant.
                if maintains_type or key in ALLTYPES_TRANSFORMED_COLS:
                    # Check that you can't "detokenize" a value
                    join_query = 'SELECT * FROM {db}.{table} WHERE {key}={fn}({value})'.format(
                        db=DB, table=ALLTYPES_TABLE, fn=PRIVACY_FN, key=key, value=val)
                    self._validate_query(ctx, conn, join_query,
                                         equal_across_queries=False, equal_across_users=False, has_results=False)

                    join_query = 'SELECT * FROM {db}.{table} WHERE {key}={fn}({value}, signed_user(false))'.format(
                        db=DB, table=ALLTYPES_PRIVACY_SHARED_VIEW, fn=PRIVACY_FN, key=key, value=val)
                    self._validate_query(ctx, conn, join_query,
                                         equal_across_queries=False, equal_across_users=False, has_results=False)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_nfp_ref_tokenize(self):
        self._test_fn('nfp_ref_tokenize', maintains_ref_integrity=True, stable=True, maintains_type=False)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_nfp_noref_tokenize(self):
        self._test_fn('nfp_noref_tokenize', maintains_ref_integrity=False, stable=True, maintains_type=False)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_fp_ref_tokenize(self):
        self._test_fn('fp_ref_tokenize', maintains_ref_integrity=True, stable=True, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_fp_noref_tokenize(self):
        self._test_fn('fp_noref_tokenize', maintains_ref_integrity=False, stable=True, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_fpt(self):
        self._test_fn('fpt', maintains_ref_integrity=False, stable=True, maintains_type=True)

    def test_fpt_ref(self):
        self._test_fn('fpt_ref', maintains_ref_integrity=True, stable=True, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_tokenize_no_ref(self):
        self._test_fn('tokenize_no_ref', maintains_ref_integrity=False, stable=True, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_tokenize(self):
        self._test_fn('tokenize', maintains_ref_integrity=True, stable=True, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_fp_random(self):
        self._test_fn('fp_random', maintains_ref_integrity=False, stable=False, maintains_type=True)

    @unittest.skipIf(common.test_level_lt('checkin'), "Skipping below checkin test level")
    def test_nfp_random(self):
        self._test_fn('fp_random', maintains_ref_integrity=False, stable=False, maintains_type=False)

if __name__ == "__main__":
    unittest.main()



#import unittest
import os
import sys
from okera.tests import pycerebro_test_common as common
sys.path.append(os.path.join(os.environ.get("OKERA_HOME"), "integration", "catalog"))
# pylint: disable=wrong-import-position
import catalog_okera
import okera_catalog
from test_catalog import mock_catalog_cls

PORT = int(os.environ['ODAS_TEST_PORT_PLANNER_THRIFT'])
HOST = os.environ['ODAS_TEST_HOST']
TOKEN = "root"
LOG_DIR = os.path.join(os.environ.get("OKERA_HOME"), "logs")

# Config dictionary for all sync tests
CONFIGS = {
    'log_directory': LOG_DIR,
    'okera_host': HOST,
    'okera_port': PORT,
    'okera_token': TOKEN,
    'catalog_objects': {
        'databases': [],
        'tables': []},
    'logging_level': "debug",
    'mapped_catalog_attributes':
    [{'name': "Catalog Classification",
      'okera_namespace': "catalog_classification"},
     {'name': "Catalog Tag",
      'okera_namespace': "catalog_tag"}],
    'sync_descriptions': True
}

# Test database name
TEST_DB = "mock_db"
TEST_DB2 = "mock_db2"

# Test table names
TRANSACTIONS_TABLE = "transactions"
TRANSACTIONS2_TABLE = "transactions2"
USERS_TABLE = "users"
USERS2_TABLE = "users2"

# DDL statements to reset Okera test environment
DB_DDL = "CREATE DATABASE IF NOT EXISTS %s;"

TRANSACTIONS_DDL = """
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
    ordernumber INT,
    quantityordered INT,
    priceeach INT,
    orderlinenumber INT,
    sales INT,
    orderdate STRING,
    status STRING,
    qtr_id INT,
    month_id INT,
    year_id INT,
    productline STRING,
    msrp INT,
    productcode STRING,
    customername STRING,
    phone STRING,
    address STRING,
    postalcode STRING,
    country STRING,
    territory STRING,
    ipaddress STRING,
    contactname STRING,
    dealsize STRING, rep STRING
)
LOCATION 's3a://okera-datalake/collibra';"""

USERS_DDL = """
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
    uid STRING,
    dob STRING,
    gender STRING,
    ccn STRING
)
LOCATION 's3a://okera-datalake/collibra';"""

DROP_DDL = "DROP DATABASE IF EXISTS %s CASCADE;"

# Add description table property to table DDL
def add_description(ddl, description):
    ddl = ddl[:-1]
    desc_ddl = " TBLPROPERTIES ('comment'='%s');"
    return ddl + desc_ddl % description

# Gets objects from Okera with tags and descriptions
# Returns list with dict for each asset, example:
# [{'name': 'example_table', 'description': 'foobar'},
#   {'name': 'col1', 'Security Classification': 'Restricted'}, ...]
def get_okera_table(db, table_name, conn):
    table = conn.list_datasets(db=db, name=table_name)[0]

    return parse_okera_object(table)

def get_okera_database(db, conn):
    objects = [{"name": db, "description": None}]

    db_tables = conn.list_datasets(db=db)

    for table in db_tables:
        objects = objects + parse_okera_object(table)

    return objects

def parse_okera_object(okera_object):
    objects = []

    full_name = '.'.join([okera_object.db[0], okera_object.name])
    tbl = {"name": full_name, "description": okera_object.description}

    if okera_object.attribute_values:
        for tbl_tag in okera_object.attribute_values:
            tbl[tbl_tag.attribute.attribute_namespace] = tbl_tag.attribute.key

    objects.append(tbl)

    for column in okera_object.schema.cols:
        col = {"name": '.'.join([full_name, column.name])}
        if column.attribute_values:
            for col_tag in column.attribute_values:
                col[col_tag.attribute.attribute_namespace] = col_tag.attribute.key

        col["description"] = column.comment
        if column.comment is None:
            col["description"] = None

        objects.append(col)

    return objects

# Gets objects from the mock catalog with tags and descriptions
# Returns list with dict for each asset, example:
# [{'name': 'example_table', 'description': 'foobar'},
#   {'name': 'col1', 'Security Classification': 'Restricted'}, ...]
def get_catalog_table(mock_catalog, table_name):
    catalog_object = mock_catalog.get_table('.'.join([TEST_DB, table_name]))
    catalog_objects = [parse_catalog_object(catalog_object)]
    if catalog_object.children:
        for child in catalog_object.children:
            catalog_objects.append(parse_catalog_object(child))

    return catalog_objects

def parse_catalog_object(catalog_object):
    obj = {"name": catalog_object.name.full_name,
           "description": catalog_object.attributes['description']}

    if catalog_object.attributes['tags']:
        for tag in catalog_object.attributes['tags']:
            obj[tag.namespace] = tag.key

    return obj

# Sync mock catalog table attributes to Okera
#
# Current test attributes:
# - `transactions`: Catalog Classification = "restricted",
#                   Description = "Mock dataset sync test."
# - `transactions.phone`: Catalog Tag = "phone", PII = "phone_number"
# - `transactions.address`: Description = "Home address", PII = "address"
# - `transactions.sales`: Description = "Number of sales"
# - `users`: Description = "Mock dataset sync test.",
#            Catalog Classification = "restricted"
# - `users.ccn`: Catalog Classification = "confidential"
# - `users.dob`: Description = "Date of birth"
#
# Reset Okera test environment,
# create database `mock_db` in Okera
# create dataset `transactions` and `users` in Okera,
# sync attributes from mock catalog to Okera
def test_table_okera():
    ctx = common.get_test_context()
    ctx.enable_token_auth(token_str=TOKEN)

    with common.get_planner(ctx, host=HOST, port=PORT) as conn:
        mock_catalog = mock_catalog_cls.MockCatalog(CONFIGS)
        # Clean up Okera environment by deleting database
        conn.execute_ddl(DROP_DDL % TEST_DB)
        # Create new database and `transactions` dataset in Okera
        conn.execute_ddl(DB_DDL % TEST_DB)
        users_ddl = add_description(
            USERS_DDL % (TEST_DB, USERS_TABLE), "Mock catalog sync dataset description.")
        conn.execute_ddl(users_ddl)
        conn.execute_ddl(TRANSACTIONS_DDL % (TEST_DB, TRANSACTIONS_TABLE))

        # Set sync objects in config dictionary
        CONFIGS['catalog_objects']['tables'] = [
            '.'.join([TEST_DB, USERS_TABLE]), '.'.join([TEST_DB, TRANSACTIONS_TABLE])]

        # Run catalog to Okera sync
        catalog_okera.run_sync("mock", CONFIGS)

        # Get users tables
        catalog_users = get_catalog_table(mock_catalog, USERS_TABLE)
        okera_users = get_okera_table(TEST_DB, USERS_TABLE, conn)

        # Get transactions tables
        catalog_transactions = get_catalog_table(mock_catalog, TRANSACTIONS_TABLE)
        okera_transactions = get_okera_table(TEST_DB, TRANSACTIONS_TABLE, conn)

        # Compare list of dicts from Okera and mock catalog to ensure sync was successful
        # The table and each of its columns have a dict that contains the assets name
        # and any attributes it has, example:
        # catalog_objects = [
        #   {'name': 'example_table', 'description': 'foobar'},
        #   {'name': 'col1', 'security Classification': 'restricted'},
        #   {'name': 'col2'}]
        # okera_objects = [
        #   {'name': 'example_table', 'description': 'foobar'},
        #   {'name': 'col1', 'security classification': 'restricted'},
        #   {'name': 'col2'}]
        for obj in catalog_users:
            assert obj in okera_users

        for obj in catalog_transactions:
            assert obj in okera_transactions

# Sync mock catalog database attributes to Okera
#
# Current test attributes:
# same as test_table_okera()
#
# Reset Okera test environment,
# create database `mock_db` in Okera
# create dataset `transactions` and `users` in Okera,
# sync attributes from mock catalog to Okera
def test_database_okera():
    ctx = common.get_test_context()
    ctx.enable_token_auth(token_str=TOKEN)

    with common.get_planner(ctx, host=HOST, port=PORT) as conn:
        mock_catalog = mock_catalog_cls.MockCatalog(CONFIGS)
        # Clean up Okera environment by deleting database
        conn.execute_ddl(DROP_DDL % TEST_DB)
        # Create new database and `transactions` dataset in Okera
        conn.execute_ddl(DB_DDL % TEST_DB)
        users_ddl = add_description(
            USERS_DDL % (TEST_DB, USERS_TABLE), "Mock catalog sync dataset description.")
        conn.execute_ddl(users_ddl)
        conn.execute_ddl(TRANSACTIONS_DDL % (TEST_DB, TRANSACTIONS_TABLE))

        # Set sync objects in config dictionary
        CONFIGS['catalog_objects']['databases'].append(TEST_DB)
        CONFIGS['catalog_objects']['tables'] = []

        # Run catalog to Okera sync
        catalog_okera.run_sync("mock", CONFIGS)

        catalog_object = mock_catalog.get_database(TEST_DB)
        catalog_objects = [parse_catalog_object(catalog_object)]
        if catalog_object.children:
            for table in catalog_object.children:
                catalog_objects.append(parse_catalog_object(table))
                if table.children:
                    for column in table.children:
                        catalog_objects.append(parse_catalog_object(column))

        okera_objects = get_okera_database(TEST_DB, conn)

        for obj in catalog_objects:
            assert obj in okera_objects

# Sync mock catalog table attributes to Okera without description
#
# Current test attributes:
# same as test_table_okera() but without descriptions
#
# Reset Okera test environment,
# create database `mock_db` in Okera
# create dataset `users` in Okera,
# sync attributes from mock catalog to Okera
def test_disable_description():
    ctx = common.get_test_context()
    ctx.enable_token_auth(token_str=TOKEN)

    with common.get_planner(ctx, host=HOST, port=PORT) as conn:
        mock_catalog = mock_catalog_cls.MockCatalog(CONFIGS)
        # Clean up Okera environment by deleting database
        conn.execute_ddl(DROP_DDL % TEST_DB)
        # Create new database and `transactions` dataset in Okera
        conn.execute_ddl(DB_DDL % TEST_DB)
        conn.execute_ddl(USERS_DDL % (TEST_DB, USERS_TABLE))

        # Disable description sync in config dictionary
        CONFIGS['sync_descriptions'] = False

        # Set sync objects in config dictionary
        CONFIGS['catalog_objects']['tables'] = ['.'.join([TEST_DB, USERS_TABLE])]

        # Run catalog to Okera sync
        catalog_okera.run_sync("mock", CONFIGS)

        # Get users tables
        catalog_users = get_catalog_table(mock_catalog, USERS_TABLE)
        okera_users = get_okera_table(TEST_DB, USERS_TABLE, conn)

        # Check if Okera descriptions are none
        for obj in okera_users:
            assert obj["description"] is None

        # Check if the objects and their other attributes were synced correctly
        for obj in catalog_users:
            assert any("name" in obj for obj in okera_users)
            for attr in CONFIGS['mapped_catalog_attributes']:
                if obj.get(attr['okera_namespace']):
                    assert any(attr['okera_namespace'] in obj for obj in okera_users)

# Sync Okera table to mock catalog
#
# Current test attributes:
# - `users`: Description = "Mock catalog sync dataset description.")
#
# Reset Okera test environment,
# create database `mock_db` in Okera
# create dataset `transactions` and `users` in Okera,
# sync tables from Okera to mock catalog
def test_okera_table():
    ctx = common.get_test_context()
    ctx.enable_token_auth(token_str=TOKEN)

    with common.get_planner(ctx, host=HOST, port=PORT) as conn:
        mock_catalog = mock_catalog_cls.MockCatalog(CONFIGS)
        # Clean up Okera environment by deleting database
        conn.execute_ddl(DROP_DDL % TEST_DB)
        # Create new database and `transactions` dataset in Okera
        conn.execute_ddl(DB_DDL % TEST_DB)
        users_ddl = add_description(
            USERS_DDL % (TEST_DB, USERS2_TABLE), "Mock catalog sync dataset description.")
        conn.execute_ddl(users_ddl)
        conn.execute_ddl(TRANSACTIONS_DDL % (TEST_DB, TRANSACTIONS2_TABLE))

        # Set sync objects in config dictionary
        CONFIGS['catalog_objects']['databases'] = []
        CONFIGS['catalog_objects']['tables'] = [
            '.'.join([TEST_DB, USERS2_TABLE]), '.'.join([TEST_DB, TRANSACTIONS2_TABLE])]


        # Run catalog to Okera sync
        okera_catalog.run_sync("mock", CONFIGS)

        # Get transactions tables
        catalog_transactions = get_catalog_table(mock_catalog, TRANSACTIONS2_TABLE)
        okera_transactions = get_okera_table(TEST_DB, TRANSACTIONS2_TABLE, conn)

        # Get users tables
        catalog_users = get_catalog_table(mock_catalog, USERS2_TABLE)
        okera_users = get_okera_table(TEST_DB, USERS2_TABLE, conn)

        for obj in catalog_transactions:
            assert obj in okera_transactions

        for obj in catalog_users:
            assert obj in okera_users

# Sync Okera database to mock catalog
#
# Current test attributes:
# - `users`: Description = "Mock catalog sync dataset description.")
#
# Reset Okera test environment,
# create database `mock_db` in Okera
# create dataset `transactions` and `users` in Okera,
# sync database with tables from Okera to mock catalog
def test_okera_database():
    ctx = common.get_test_context()
    ctx.enable_token_auth(token_str=TOKEN)

    with common.get_planner(ctx, host=HOST, port=PORT) as conn:
        mock_catalog = mock_catalog_cls.MockCatalog(CONFIGS)
        # Clean up Okera environment by deleting database
        conn.execute_ddl(DROP_DDL % TEST_DB2)
        # Create new database and `transactions` dataset in Okera
        conn.execute_ddl(DB_DDL % TEST_DB2)
        users_ddl = add_description(
            USERS_DDL % (TEST_DB2, USERS_TABLE), "Mock catalog sync dataset description.")
        conn.execute_ddl(users_ddl)
        conn.execute_ddl(TRANSACTIONS_DDL % (TEST_DB2, TRANSACTIONS_TABLE))

        # Set sync objects in config dictionary
        CONFIGS['catalog_objects']['databases'].append(TEST_DB2)
        CONFIGS['catalog_objects']['tables'] = []

        # Run catalog to Okera sync
        okera_catalog.run_sync("mock", CONFIGS)

        catalog_object = mock_catalog.get_database(TEST_DB2)
        catalog_objects = [parse_catalog_object(catalog_object)]
        if catalog_object.children:
            for table in catalog_object.children:
                catalog_objects.append(parse_catalog_object(table))
                if table.children:
                    for column in table.children:
                        catalog_objects.append(parse_catalog_object(column))

        okera_objects = get_okera_database(TEST_DB2, conn)

        for obj in catalog_objects:
            assert obj in okera_objects



#pylint: skip-file

from functools import wraps
from warnings import warn

import okera._pure_sasl_mechanisms as mech_mod
from okera._pure_sasl import SASLError, QOP


def _require_mech(f):
    """
    A utility decorator that ensures a mechanism has been chosen.
    """
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        if not self._chosen_mech:
            raise SASLError("A mechanism has not been chosen yet")
        return f(self, *args, **kwargs)

    return wrapped


class SASLClient(object):
    """
    A SASL client for one negotiation with a SASL server.

    An new instance of this is typically created when establishing a connection
    with a SASL server. The SASL mechanism may be chosen at creation time or
    set later after a list of server-supported mechanisms has been discovered
    using `:method:choose_mechanism()`.

    Instances of this class do not directly communicate with the server (as
    the communication protocol differs from service to service), but rather
    relies on challenges from the server to be passed in and processed.  When
    initiating a new connection, this is done by passing challenges from the
    server to the :meth:`process()` method until the server indicates that
    SASL negotiation has completed and the `complete` attribute of the instance
    is set to ``True``.

    After the initial negotiation is complete, communication between the
    server and client may need to be processed, depending on the 'quality of
    service', which may call for signatures or encryption. To handle this,
    messages from the server should be passed through :meth:`unwrap()`, and
    messages outbound from the client to server should be passed through
    :meth:`wrap()`.

    Example usage::

        >>> from okera._pure_sasl_client import SASLClient
        >>>
        >>> sasl = SASLClient('somehost2', 'customprotocol')
        >>> conn = get_connection_to('somehost2')
        >>> available_mechs = conn.get_mechanisms()
        >>> sasl.choose_mechanism(available_mechs, allow_anonymous=False)
        >>> while True:
        ...     status, challenge = conn.get_challenge()
        ...     if status == 'COMPLETE':
        ...         break
        ...     elif status == 'OK':
        ...         response = sasl.process(challenge)
        ...         conn.send_response(response)
        ...     else:
        ...         raise Exception(status)
        ...
        >>> if not sasl.complete:
        ...     raise Exception("SASL negotiation did not complete")
        >>>
        >>> # begin normal communication
        >>> encoded = conn.fetch_data()
        >>> decoded = sasl.unwrap(encoded)
        >>> response = process_data(decoded)
        >>> conn.send_data(sasl.wrap(response))
    """

    def __init__(self, host, service=None, mechanism=None, authorization_id=None,
                 callback=None, qops=QOP.all, mutual_auth=False, max_buffer=65536,
                 **mechanism_props):
        """
        `host` is the name of the SASL server, typically an FQDN, and `service` is
        usually the name of the protocol, such as `imap` or `http`.

        `mechanism` may be the string name of a mechanism to use, like
        'PLAIN' or 'GSSAPI'.  If left as ``None``, :meth:`choose_mechanism`
        must be used with a list of mechanisms that the server supports before
        `process()` can be used.

        Optionally, an `authorization_id` may be set if the mechanism and protocol
        support authorization.

        The allowed quality of protection (QoP) choices may be set with the `qops`
        parameter, which should be an iterable of allowed options. Valid options
        include 'auth' for no protection, 'auth-int' for integrity protection,
        and 'auth-conf' for confidentiality protection. The strongest of these
        that the server also supports will be chosen automatically.  If the
        server does not support any of these choices, a
        :exc:`SASLProtocolException` will be raised.

        If the chosen mechanism supports mutual authentication, which is
        authentication of the server by the client, this may be set to
        ``True`` to ensure that mutual authentication is performed.

        A max buffer size may be set with `max_buffer`.  If a max buffer size
        is also set during negotiation by the server, the min of these two
        values will be used.

        Any other mechanism-specific properties may be set with
        `**mechanism_props` and will automatically be passed in to the
        mechanism's constructor.  If any properties are required by the
        mechanism during the course of negotiation have not been passed in
        via `**mechanism_props`, the function passed in here as the `callback`
        argument will be called with one argument, the name of the required
        property.  The `callback` function should return a value for that
        property.
        """
        self.host = host
        self.service = service
        self.authorization_id = authorization_id
        self.mechanism = mechanism
        self.callback = callback
        self.qops = set(qops)
        self.mutual_auth = mutual_auth
        self.max_buffer = max_buffer

        self._mech_props = mechanism_props
        if self.mechanism is not None:
            try:
                mech_class = mech_mod.mechanisms[mechanism]
            except KeyError:
                gssapi = mech_mod.GSSAPIMechanism.name
                if mechanism == gssapi and not mech_mod.have_kerberos:
                    raise SASLError('kerberos module not installed, {0} '
                                    'unavailable'.format(gssapi))
                else:
                    raise SASLError('Unknown mechanism {0}'.format(mechanism))
            self._chosen_mech = mech_class(self, **self._mech_props)
        else:
            self._chosen_mech = None

    @_require_mech
    def process(self, challenge=None):
        """
        Process a challenge from the server during SASL negotiation.
        A response will be returned which should typically be sent to the
        server to answer the challenge.

        With some mechanisms and protocols, `process()` should be called
        with a `challenge` of ``None`` to generate the first message
        to be sent to the server.
        """
        return self._chosen_mech.process(challenge)

    @_require_mech
    def wrap(self, outgoing):
        """
        Wrap an outgoing message intended for the SASL server. Depending
        on the negotiated quality of protection, this may result in the
        message being signed, encrypted, or left unaltered.
        """
        return self._chosen_mech.wrap(outgoing)

    @_require_mech
    def unwrap(self, incoming):
        """
        Unwrap a message from the SASL server. Depending on the negotiated
        quality of protection, this may check a signature, decrypt the message,
        or leave the message unaltered.
        """
        return self._chosen_mech.unwrap(incoming)

    @property
    def complete(self):
        """
        Check to see if SASL negotiation has completed successfully, including
        a mutual authentication check if the chosen mechanism supports that and
        mutual authentication was requested via the `mutual_auth` parameter
        for the `SASLClient` constructor.
        """
        if not self._chosen_mech:
            raise SASLError("A mechanism has not been chosen yet")
        return self._chosen_mech.complete

    @_require_mech
    def dispose(self):
        """
        Clear all sensitive data, such as passwords.
        """
        self._chosen_mech.dispose()

    @property
    @_require_mech
    def qop(self):
        return self._chosen_mech.qop

    def choose_mechanism(self, mechanism_choices, allow_anonymous=True,
                         allow_plaintext=True, allow_active=True,
                         allow_dictionary=True):
        """
        Choose a mechanism from a list of mechanisms based on security
        scores for mechanisms and required properties of the mechanism.

        If `allow_anonymous` is ``False``, mechanisms that allow anonymous
        authentication will not be considered.

        If `allow_plaintext` is ``False``, mechanisms that transmit
        sensitive information in plaintext (and are thus susceptible to
        passive listening attacks) will not be considered.

        If `allow_active` is ``False``, mechanisms that are susceptible
        to active non-dictionary attacks (MITM, injection) will not be
        considered.

        If `allow_dictionary` is ``False, mechanisms that are susceptible
        to passive dictionary attacks will not be considered.
        """
        gssapi = mech_mod.GSSAPIMechanism.name
        if gssapi in mechanism_choices and not mech_mod.have_kerberos:
            warn('kerberos module not installed, {0} will be ignored'.format(
                 gssapi), SASLWarning)

        candidates = [mech_mod.mechanisms[choice]
                      for choice in mechanism_choices
                      if choice in mech_mod.mechanisms]

        if not allow_anonymous:
            candidates = [m for m in candidates if not m.allows_anonymous]
        if not allow_plaintext:
            candidates = [m for m in candidates if not m.uses_plaintext]
        if not allow_active:
            candidates = [m for m in candidates if m.active_safe]
        if not allow_dictionary:
            candidates = [m for m in candidates if m.dictionary_safe]

        if not candidates:
            raise SASLError("None of the mechanisms listed meet all "
                            "required properties")

        # Pick the best mechanism based on its security score
        mech_class = max(candidates, key=lambda mech: mech.score)
        self.mechanism = mech_class.name
        self._chosen_mech = mech_class(self, **self._mech_props)



# Copyright 2019 Okera Inc. All Rights Reserved.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines
#
# Tests for Mindsphere API
#

import copy
import json
import os
import unittest

from okera._thrift_api import TAuthorizeQueryParams
from okera._thrift_api import TAuthorizeQueryClient
from okera._thrift_api import TCompoundOp
from okera._thrift_api import TErrorCode
from okera._thrift_api import TRecordServiceException

from okera.tests import pycerebro_test_common as common

MEASURE = common.get_env_var('MEASURE', bool, False)
ENABLE_OR = common.get_env_var('ENABLE_OR', bool, False)
# We don't enable grants into complex types by default. Need to
# turn on server side with flag.
# Need to run: start-cdas-container.sh \
#     -enable-complex-types-tags \
#     -env "CASCADE_NESTED_TYPES_CHECK=true"
COMPLEX_TYPES = common.get_env_var('COMPLEX_TYPES', bool, False)

ASSETS_LOC = 's3://cerebrodata-test/mindsphere/assets/example.json'
TEST_DB = 'mindsphere_test'
TEST_ROLE = 'mindsphere_role'
TEST_USER = 'mindsphere_user'

# FIXME: original record has unicode for streetAddress which gets decoded wrong. This
# is very likely a test issue.
ASSET_FULL_RECORD = {
  "name": "Millenium Falcon",
  "externalId": "SN 123456-123-123456",
  "description": "The ship of Han Solo and Chewbacca",
  "location": {
    "country": "Austria",
    "region": "Tyrol",
    "locality": "Innsbruck",
    "streetAddress": "Industriest 21 A/II",
    "postalCode": "6020",
    "longitude": 53.5125546,
    "latitude": 9.9763411
  },
  "variables": [
    {
      "name": "color",
      "value": "yellow"
    }
  ],
  "aspects": [
    {
      "name": "astroDroid",
      "variables": [
        {
          "name": "color",
          "value": "yellow"
        }
      ]
    }
  ],
  "fileAssignments": [
    {
      "key": "logo_small",
      "fileId": "c27a28b6eb16b507fabc11e75da8b4ce"
    }
  ],
  "typeId": "mdsp.spaceship",
  "parentId": "c27a28b6eb16b507fabc11e75da8b4ce",
  "timezone": "Europe/Berlin",
  "twinType": "performance",
  "tenantId": "mdsp",
  "subTenant": "UnkarPlutt Inc.",
  "t2Tenant": "DEPRECATED: use subTenant instead",
  "assetId": "c27a28b6eb16b507fabc11e75da8b4ce",
  "locks": [
    {
      "id": "c27a28b6eb16b507fabc11e75da8b4ce",
      "service": "AgentManagement",
      "reason": "Agent is onboarded, cannot delete asset until offboard finished",
      "reasonCode": "agentmanagement.agent.onboarded"
    }
  ],
  "deleted": "2020-12-03T06:27:43.126Z",
  "sharing": {
    "modes": [
      "SHARER"
    ]
  },
  "etag": 1,
  "hierarchyPath": [
    {
      "assetId": "c27a28b6eb16b507fabc11e75da8b4ce",
      "name": "Millenium Falcon's parent"
    }
  ]
}

# Make some derived records
ASSET_RECORD_GERMANY = copy.deepcopy(ASSET_FULL_RECORD)
ASSET_RECORD_GERMANY['location']['country'] = 'Germany'
ASSET_RECORD_GERMANY['name'] = 'Millenium Falcon2';
ASSET_RECORD_GERMANY['hierarchyPath'].append({"assetId": "abcd", "name": "p2"})
ASSET_RECORD_GERMANY['aspects'][0]['variables'].append({"name": "k", "value": "mdsp"})
#print(json.dumps(ASSET_RECORD_GERMANY, indent=2))

class MindsphereApiTest(common.TestBase):
    @classmethod
    def setUpClass(cls):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % TEST_DB)
            conn.execute_ddl("CREATE DATABASE %s" % TEST_DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.assets
                LIKE JSON '%s'
                WITH SERDEPROPERTIES('new.line.json'='false') STORED AS JSON
                LOCATION '%s'
                TBLPROPERTIES ('okera.catalog.cache-metadata'='true')
                ''' % (TEST_DB, ASSETS_LOC, ASSETS_LOC))

    def _authorize_query(self, conn, db, tbl, records, user=None,
                        client=TAuthorizeQueryClient.REST_API,
                        return_records=True, skip_result=False,
                        user_filter=None):
        request = TAuthorizeQueryParams()
        request.db = [db]
        request.dataset = tbl
        request.records = records
        request.requesting_user = user
        request.client = client
        request.return_records = return_records
        request.request_filter = user_filter
        result = conn.service.client.AuthorizeQuery(request)
        if not skip_result and result.result_records:
            result_records = []
            for r in result.result_records:
              result_records.append(json.loads(r))
            return result.filter, result_records, result
        else:
            return result.filter, None, result

    def _authorize_sql(self, conn, sql, user=None,
                       user_filter=None,
                       client=TAuthorizeQueryClient.REST_API):
        request = TAuthorizeQueryParams()
        request.sql = sql
        request.requesting_user = user
        request.client = client
        request.request_filter = user_filter
        result = conn.service.client.AuthorizeQuery(request)
        return ' '.join(result.result_sql.split())

    def _authorize_assets(self, label, conn, user, data,
                          user_filter=None, measure=False, iters=2000,
                          client = TAuthorizeQueryClient.REST_API):
        input_records = None
        if data is not None:
            input_records = []
            for r in data:
                input_records.append(json.dumps(r))

        def get():
            return self._authorize_query(conn, TEST_DB, 'assets', input_records,
                                         user=user, user_filter=user_filter,
                                         skip_result=True, client=client)
        if measure or MEASURE:
            print()
            print('---------------------------- %s --------------------------' % label)
            common.measure_latency(iters, get)
        return self._authorize_query(conn, TEST_DB, 'assets', input_records, user=user,
                                     user_filter=user_filter, client=client)

    def test_admin(self):
        """ The tests that verify more records spend a lot of time in pytest. """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            """
            ---------------------------- admin --------------------------
            Iterations 2000
            Mean 0.35735225677490234 ms
            50%: 0.3466606140136719 ms
            90%: 0.3914833068847656 ms
            95%: 0.40459632873535156 ms
            99%: 0.4324913024902344 ms
            99.5%: 0.44798851013183594 ms
            99.9%: 0.5857944488525391 ms
            """
            filter, records, _ = self._authorize_assets(
                "admin", conn, None, [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertTrue(filter is None)
            self.assertEqual(records, [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])

            """
            ------------------------------ admin100 ----------------------------
            Iterations 2000
            Mean 1.7077339887619019 ms
            50%: 1.6613006591796875 ms
            90%: 1.7123222351074219 ms
            95%: 1.7268657684326172 ms
            99%: 3.3609867095947266 ms
            99.5%: 4.5528411865234375 ms
            99.9%: 10.724544525146484 ms
            """
            records_100 = []
            for _ in range(0, 100):
                records_100.append(ASSET_FULL_RECORD)
            filter, records, _ = self._authorize_assets(
                "admin100", conn, None, records_100)
            self.assertTrue(filter is None)
            self.assertEqual(records, records_100)

            """
            ------------------------------ admin2000 ----------------------------
            Iterations 300
            Mean 30.05538543065389 ms
            50%: 29.486656188964844 ms
            90%: 31.653881072998047 ms
            95%: 31.9976806640625 ms
            99%: 38.814544677734375 ms
            99.5%: 40.86470603942871 ms
            99.9%: 44.52395439147949 ms
            """
            records_2000 = []
            for _ in range(0, 2000):
                records_2000.append(ASSET_FULL_RECORD)
            filter, records, _ = self._authorize_assets(
                "admin2000", conn, None, records_2000, iters=300)
            self.assertTrue(filter is None)
            self.assertEqual(records, records_2000)

            """
            ------------------------------ admin10000 ----------------------------
            Iterations 100
            Mean 159.57778453826904 ms
            50%: 154.77943420410156 ms
            90%: 160.01248359680176 ms
            95%: 195.74332237243652 ms
            99%: 255.62667846679688 ms
            99.5%: 255.62667846679688 ms
            99.9%: 255.62667846679688 ms
            """
            records_10000 = []
            for _ in range(0, 10000):
                records_10000.append(ASSET_FULL_RECORD)
            filter, records, _ = self._authorize_assets(
                "admin10000", conn, None, records_10000, iters=100)
            self.assertTrue(filter is None)
            self.assertEqual(records, records_10000)

            """
            ------------------------------ admin25000 ----------------------------
            Iterations 50
            Mean 465.39061546325684 ms
            50%: 510.6058120727539 ms
            90%: 527.1403789520264 ms
            95%: 528.7942886352539 ms
            99%: 556.159496307373 ms
            99.5%: 556.159496307373 ms
            99.9%: 556.159496307373 ms
            """
            records_25000 = []
            for _ in range(0, 25000):
                records_25000.append(ASSET_FULL_RECORD)
            filter, records, _ = self._authorize_assets(
                "admin25000", conn, None, records_25000, iters=50)
            self.assertTrue(filter is None)
            self.assertEqual(records, records_25000)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets")

    def test_name_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "name = 'Millenium Falcon'", TEST_ROLE))

            """
            ------------------------------ name filter ----------------------------
            Iterations 2000
            Mean 0.781161904335022 ms
            50%: 0.7805824279785156 ms
            90%: 0.8177757263183594 ms
            95%: 0.8411407470703125 ms
            99%: 1.1434555053710938 ms
            99.5%: 1.1930465698242188 ms
            99.9%: 3.467082977294922 ms
            """
            filter, records, res = self._authorize_assets(
                "name filter", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "name = 'Millenium Falcon'")
            self.assertEqual(records, [ASSET_FULL_RECORD])
            self.assertTrue(res.column_access is None)
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "name = 'Millenium Falcon'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            """
            ------------------------------ name filter 100 ----------------------------
            Iterations 2000
            Mean 3.7372668981552124 ms
            50%: 3.6094188690185547 ms
            90%: 3.6880970001220703 ms
            95%: 5.043506622314453 ms
            99%: 5.868673324584961 ms
            99.5%: 8.349895477294922 ms
            99.9%: 12.394905090332031 ms
            """
            records_100 = []
            for _ in range(0, 100):
                records_100.append(ASSET_FULL_RECORD)
            filter, records, _ = self._authorize_assets(
                "name filter 100", conn, TEST_USER, records_100)
            self.assertEqual(filter, "name = 'Millenium Falcon'")
            self.assertEqual(records, records_100)

            """
            ------------------------------ name filter 2000 ----------------------------
            Iterations 300
            Mean 37.56293932596842 ms
            50%: 35.05730628967285 ms
            90%: 43.69378089904785 ms
            95%: 47.528982162475586 ms
            99%: 58.081865310668945 ms
            99.5%: 62.833309173583984 ms
            99.9%: 157.52625465393066 ms
            """
            records_2000 = []
            for _ in range(0, 2000):
                records_2000.append(ASSET_RECORD_GERMANY)
            filter, records, _ = self._authorize_assets(
                "name filter 2000", conn, TEST_USER, records_2000, iters=300)
            self.assertEqual(filter, "name = 'Millenium Falcon'")
            self.assertTrue(records is None)

            """
            ------------------------------ name filter 10000 ----------------------------
            Iterations 100
            Mean 454.8932361602783 ms
            50%: 529.1121006011963 ms
            90%: 583.2486152648926 ms
            95%: 588.5665416717529 ms
            99%: 596.2774753570557 ms
            99.5%: 596.2774753570557 ms
            99.9%: 596.2774753570557 ms
            """
            records_10000 = []
            for _ in range(0, 5000):
                records_10000.append(ASSET_FULL_RECORD)
                records_10000.append(ASSET_RECORD_GERMANY)
            filter, records, _ = self._authorize_assets(
                "name filter 10000", conn, TEST_USER, records_10000, iters=100)
            self.assertEqual(filter, "name = 'Millenium Falcon'")
            self.assertEqual(len(records), 5000)

            # FIXME: broken, likely a batching issue
            #records_25000 = []
            #for _ in range(0, 5000):
            #    records_10000.append(ASSET_FULL_RECORD)
            #    records_10000.append(ASSET_RECORD_GERMANY)
            #    records_10000.append(ASSET_RECORD_GERMANY)
            #    records_10000.append(ASSET_RECORD_GERMANY)
            #    records_10000.append(ASSET_RECORD_GERMANY)
            #filter, records, _ = self._authorize_assets(
            #    "name filter 25000", conn, TEST_USER, records_25000, iters=30)
            #self.assertEqual(filter, "name = 'Millenium Falcon'")
            #self.assertEqual(len(records), 5000)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE name = 'Millenium Falcon'")

    def test_location_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "`location`.country = 'Austria'", TEST_ROLE))

            """
            ------------------------------ country filter ----------------------------
            Iterations 2000
            Mean 0.8221160173416138 ms
            50%: 0.7987022399902344 ms
            90%: 0.8695125579833984 ms
            95%: 0.9033679962158203 ms
            99%: 1.220703125 ms
            99.5%: 1.2383460998535156 ms
            99.9%: 3.914356231689453 ms
            """
            filter, records, res = self._authorize_assets(
                "country filter", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "`location`.country = 'Austria'")
            self.assertEqual(records, [ASSET_FULL_RECORD])
            self.assertTrue(res.column_access is None)
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'Austria'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            """
            ---------------------------- location filter 100 --------------------------
            Iterations 2000
            Mean 3.0230764150619507 ms
            50%: 2.9752254486083984 ms
            90%: 3.0531883239746094 ms
            95%: 3.080129623413086 ms
            99%: 4.962682723999023 ms
            99.5%: 5.261659622192383 ms
            99.9%: 9.050607681274414 ms
            """
            records_100 = []
            for _ in range(0, 50):
                records_100.append(ASSET_FULL_RECORD)
                records_100.append(ASSET_RECORD_GERMANY)
            filter, records, _ = self._authorize_assets(
                "location filter 100", conn, TEST_USER, records_100)
            self.assertEqual(filter, "`location`.country = 'Austria'")
            self.assertTrue(len(records), 50)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "`location`.country = 'Austria'")

    def test_dax_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "name = 'Millenium Falcon'", TEST_ROLE))

            _, _, res = self._authorize_assets(
                "dax name filter", conn, TEST_USER, [], client=TAuthorizeQueryClient.DAX)
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], '[name] = "Millenium Falcon"')
            self.assertEqual(res.filters.op, TCompoundOp.OR)

    def test_longitude_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "`location`.longitude > 50", TEST_ROLE))
            """
            ---------------------------- longitude filter --------------------------
            Iterations 2000
            Mean 0.8221020698547363 ms
            50%: 0.8089542388916016 ms
            90%: 0.8618831634521484 ms
            95%: 0.9002685546875 ms
            99%: 0.980377197265625 ms
            99.5%: 1.0075569152832031 ms
            99.9%: 2.923727035522461 ms
            """
            filter, records, _ = self._authorize_assets(
                "longitude filter", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "`location`.longitude > 50")
            self.assertEqual(records, [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "`location`.longitude > 50")

    def test_hierarchy_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "hierarchyPath.assetId = 'c27a28b6eb16b507fabc11e75da8b4ce'",
                 TEST_ROLE))

            # Filter passes
            """
            ------------------------------ hierarchy filter ----------------------------
            Iterations 2000
            Mean 0.7893334627151489 ms
            50%: 0.7739067077636719 ms
            90%: 0.8084774017333984 ms
            95%: 0.8203983306884766 ms
            99%: 0.8497238159179688 ms
            99.5%: 0.8871555328369141 ms
            99.9%: 2.8867721557617188 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchy filter", conn, TEST_USER,
                [ASSET_FULL_RECORD])
            self.assertEqual(filter,
                             "hierarchypath.assetid = 'c27a28b6eb16b507fabc11e75da8b4ce'")
            self.assertEqual(records, [ASSET_FULL_RECORD])

            # Filter passes both records
            """
            ------------------------------ hierarchy 2 filter ----------------------------
            Iterations 2000
            Mean 0.8403977155685425 ms
            50%: 0.8394718170166016 ms
            90%: 0.8664131164550781 ms
            95%: 0.8797645568847656 ms
            99%: 0.8983612060546875 ms
            99.5%: 0.9114742279052734 ms
            99.9%: 3.549337387084961 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchy 2 filter", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(records, [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])

            # Filter just on the second record
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "hierarchyPath.assetId = 'abcd'",
                 TEST_ROLE))

            # Filter fails
            """
            ------------------------------ hierarchy 0 filter ----------------------------
            Iterations 2000
            Mean 0.7549749612808228 ms
            50%: 0.7441043853759766 ms
            90%: 0.8380413055419922 ms
            95%: 0.8676052093505859 ms
            99%: 0.9095668792724609 ms
            99.5%: 0.9407997131347656 ms
            99.9%: 2.610445022583008 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchy 0 filter", conn, TEST_USER,
                [ASSET_FULL_RECORD])
            self.assertEqual(filter, "hierarchypath.assetid = 'abcd'")
            self.assertTrue(records is None)

            # Filter only passes one record
            """
            ----------------------- hierarchy partial filter ----------------------------
            Iterations 2000
            Mean 0.7380911111831665 ms
            50%: 0.7183551788330078 ms
            90%: 0.7965564727783203 ms
            95%: 0.8089542388916016 ms
            99%: 0.8356571197509766 ms
            99.5%: 0.8456707000732422 ms
            99.9%: 2.641439437866211 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchy partial filter", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(records, [ASSET_RECORD_GERMANY])

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets " +
                "WHERE hierarchypath.assetid = 'abcd'")

    def test_variables_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "aspects.variables.name = 'color' " +\
                 "AND aspects.variables.value = 'yellow'", TEST_ROLE))

            # Filter should pass
            """
            --------------------------- hierarchyPath filter --------------------------
            Iterations 2000
            Mean 0.8147140741348267 ms
            50%: 0.7791519165039062 ms
            90%: 0.8702278137207031 ms
            95%: 0.8976459503173828 ms
            99%: 0.9670257568359375 ms
            99.5%: 1.1227130889892578 ms
            99.9%: 6.104707717895508 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchyPath filter", conn, TEST_USER,
                [ASSET_FULL_RECORD])
            self.assertEqual(filter,
                             "aspects.variables.name = 'color' AND " +\
                             "aspects.variables.value = 'yellow'")
            self.assertEqual(records, [ASSET_FULL_RECORD])

            """
            ------------------ hierarchyPath filter partial --------------------------
            Iterations 2000
            Mean 0.8471413850784302 ms
            50%: 0.8230209350585938 ms
            90%: 0.9124279022216797 ms
            95%: 0.9305477142333984 ms
            99%: 0.9970664978027344 ms
            99.5%: 1.0478496551513672 ms
            99.9%: 3.6346912384033203 ms
            """
            filter, records, _ = self._authorize_assets(
                "hierarchyPath filter partial", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(records, [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])

            # Test some filters that mix collection types and not. This, if naively
            # constructed as a table would be jagged (collection types are longer).
            """
            ---------------- filter: aspects.variable.value = tenantId ----------------
            Iterations 2000
            Mean 0.8313227891921997 ms
            50%: 0.8170604705810547 ms
            90%: 0.8463859558105469 ms
            95%: 0.8561611175537109 ms
            99%: 0.8795261383056641 ms
            99.5%: 0.90789794921875 ms
            99.9%: 3.557920455932617 ms
            """
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "aspects.variables.value = tenantId", TEST_ROLE))
            filter, records, _ = self._authorize_assets(
                "filter: aspects.variable.value = tenantId", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "aspects.variables.value = tenantid")
            self.assertEqual(records, [ASSET_RECORD_GERMANY])

            """
            Iterations 2000
            Mean 0.8139365911483765 ms
            50%: 0.7936954498291016 ms
            90%: 0.8418560028076172 ms
            95%: 0.8752346038818359 ms
            99%: 0.9112358093261719 ms
            99.5%: 0.9334087371826172 ms
            99.9%: 3.436565399169922 ms
            """
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "aspects.variables.value = name", TEST_ROLE))
            filter, records, _ = self._authorize_assets(
                "filter: aspects.variable.value = name", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "aspects.variables.value = name")
            self.assertTrue(records is None)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "aspects.variables.value = name")

    def test_filter_folding(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "user() = name", TEST_ROLE))

            """
            ---------------------------- filter folding user --------------------------
            Iterations 2000
            Mean 0.8202756643295288 ms
            50%: 0.7958412170410156 ms
            90%: 0.8568763732910156 ms
            95%: 0.8816719055175781 ms
            99%: 0.99945068359375 ms
            99.5%: 1.0437965393066406 ms
            99.9%: 3.7946701049804688 ms
            """
            filter, records, res = self._authorize_assets(
                "filter folding user", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "name = 'mindsphere_user'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "name = 'mindsphere_user'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            self.assertTrue(records is None)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "name = 'mindsphere_user'")

            #  Test constant folding to false
            """
            ------------------------ filter folding compound 1 ------------------------
            Iterations 2000
            Mean 0.8916885852813721 ms
            50%: 0.8323192596435547 ms
            90%: 1.0120868682861328 ms
            95%: 1.2159347534179688 ms
            99%: 1.2848377227783203 ms
            99.5%: 1.329660415649414 ms
            99.9%: 2.9823780059814453 ms
            """
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "name = 'test' AND user() = 'not-a-user'", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound 1", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "FALSE")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "FALSE")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Get the authorized SQL query
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_sql(conn,
                    "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertTrue('Access to this dataset' in str(ex_ctx.exception))

            # Test constant folding to false
            """
            --------------------- filter folding compound 2 ------------------------
            Iterations 2000
            Mean 0.7768501043319702 ms
            50%: 0.7646083831787109 ms
            90%: 0.7910728454589844 ms
            95%: 0.8072853088378906 ms
            99%: 0.8356571197509766 ms
            99.5%: 0.8499622344970703 ms
            99.9%: 2.7077198028564453 ms
            """
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "user() = 'not-a-user'", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound 2", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "FALSE")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "FALSE")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Get the authorized SQL query
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_sql(conn,
                    "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertTrue('Access to this dataset' in str(ex_ctx.exception))

            # Test constant folding to true
            """
            ---------------------- filter folding compound 3 -----------------------
            Iterations 2000
            Mean 0.8791671991348267 ms
            50%: 0.8692741394042969 ms
            90%: 0.9055137634277344 ms
            95%: 0.9191036224365234 ms
            99%: 0.9796619415283203 ms
            99.5%: 1.0118484497070312 ms
            99.9%: 2.900362014770508 ms
            """
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "name = 'test' AND user() IS NOT NULL", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound 3", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(filter, "name = 'test'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "name = 'test'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "name = 'test'")

            # Try one with now(), should always be true
            """
            --------------------- filter folding compound 4 -----------------------
            Iterations 2000
            Mean 0.8819392919540405 ms
            50%: 0.8478164672851562 ms
            90%: 0.9102821350097656 ms
            95%: 0.9627342224121094 ms
            99%: 1.3670921325683594 ms
            99.5%: 1.4171600341796875 ms
            99.9%: 3.5381317138671875 ms
            """
            # TODO: test fails in jenkins
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "hour(now()) <= 24", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound 4", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            #self.assertTrue(filter is None)
            #self.assertTrue(res.filters is None)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets")

    def test_siemens_filters(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            perm = """
            `location`.`name` = '866a64ff-d9a4-4744-af13-0483edd8ddb8' AND
            name IN ('f9bf301b-d15c-4185-b4d6-8dbe20b30430',
                     'a8ccc735-c536-49cc-9d56-811fcb100162',
                     '57d423b0-efc9-449f-bdd8-b4047ea24bf4',
                     '489ca6b1-13c8-4386-9425-3da1c677b692')
            OR (
              file.assignments = '75276c8f-cb82-4630-97f7-a42a352550a9'
              AND dayname(now()) != 'anyday')
            """
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, perm, TEST_ROLE))

            # dayname should be folded out
            _, _, res = self._authorize_assets(
                "siemens filter 1", conn, TEST_USER, [])
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], ' '.join(
                '''
                  `location`.name = '866a64ff-d9a4-4744-af13-0483edd8ddb8' AND
                  name IN ('f9bf301b-d15c-4185-b4d6-8dbe20b30430',
                            'a8ccc735-c536-49cc-9d56-811fcb100162',
                            '57d423b0-efc9-449f-bdd8-b4047ea24bf4',
                            '489ca6b1-13c8-4386-9425-3da1c677b692')
                  OR
                  file.assignments = '75276c8f-cb82-4630-97f7-a42a352550a9'
                '''.split()))

            # lots of folding
            _, _, res = self._authorize_assets(
                "siemens filter 2", conn, TEST_USER, [], user_filter="name='not-there'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], ' '.join(
                '''
                file.assignments = '75276c8f-cb82-4630-97f7-a42a352550a9' AND
                name = 'not-there'
                '''.split()))

            # lots of folding
            _, _, res = self._authorize_assets(
                "siemens filter 3", conn, TEST_USER, [],
                user_filter="name='f9bf301b-d15c-4185-b4d6-8dbe20b30430'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], ' '.join(
                '''
                `location`.name = '866a64ff-d9a4-4744-af13-0483edd8ddb8' AND
                name = 'f9bf301b-d15c-4185-b4d6-8dbe20b30430'
                OR
                file.assignments = '75276c8f-cb82-4630-97f7-a42a352550a9' AND
                name = 'f9bf301b-d15c-4185-b4d6-8dbe20b30430'
                '''.split()))

    @unittest.skipIf(not ENABLE_OR, "Skipping as this requires flag to enable")
    def test_multi_grants_or(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # France OR Austria
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "`location`.country = 'Austria'", TEST_ROLE))
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "`location`.country = 'France'", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(len(res.filters.filters), 2)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'Austria'")
            self.assertEqual(res.filters.filters[1], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "`location`.country IN ('Austria', 'France')")

            # Add user filter: True, no-op
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None, user_filter='True')
            self.assertEqual(len(res.filters.filters), 2)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'Austria'")
            self.assertEqual(res.filters.filters[1], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "`location`.country IN ('Austria', 'France')")

            # Add user filter: False, overrides everything else (AND semantics)
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None, user_filter='False')
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "FALSE")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB,
                user=TEST_USER, user_filter='False')
            # FIXME
            #self.assertEqual(sql,
            #    "SELECT name FROM mindsphere_test.assets WHERE FALSE")

            # Add user filter for France, this should remove the other in clause
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None,
                user_filter="`location`.country = 'France'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Add user filter for UK, this should become false
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None,
                user_filter="`location`.country = 'UK'")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "FALSE")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            # Get the authorized SQL query

            # Add user filter for France or Austria
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None,
                user_filter="`location`.country IN ('France', 'Austria')")
            self.assertEqual(len(res.filters.filters), 2)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'Austria'")
            self.assertEqual(res.filters.filters[1], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Add user filter for France or UK, should be intersection
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER, None,
                user_filter="`location`.country IN ('France', 'UK')")
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Grant something that's always false, should just be dropped
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "user() = 'no-a-user'", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(len(res.filters.filters), 2)
            self.assertEqual(res.filters.filters[0], "`location`.country = 'Austria'")
            self.assertEqual(res.filters.filters[1], "`location`.country = 'France'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "`location`.country IN ('Austria', 'France')")

            # Grant something that's always true, no more filters
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "user() is not null", TEST_ROLE))
            filter, records, res = self._authorize_assets(
                "filter folding compound", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertTrue(res.filters is None)
            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets")

    def test_invalid_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            # Not a path in the schema
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "hierarchy.name is null", TEST_ROLE))
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_assets(
                    "filter: hierarchy.name", conn, TEST_USER,
                    [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertTrue('Could not resolve' in str(ex_ctx.exception))

            """
            ------------------------- filter: hierarchy.name ------------------------
            Iterations 2000
            Mean 0.568452000617981 ms
            50%: 0.5667209625244141 ms
            90%: 0.5910396575927734 ms
            95%: 0.6008148193359375 ms
            99%: 0.6263256072998047 ms
            99.5%: 0.6392002105712891 ms
            99.9%: 0.7848739624023438 ms
            """
            filter, _, res = self._authorize_assets(
                "filter: hierarchy.name", conn, TEST_USER, None)
            self.assertEqual(filter, 'hierarchy.name IS NULL')
            self.assertEqual(res.filters.filters[0], 'hierarchy.name IS NULL')
            self.assertEqual(res.filters.op, TCompoundOp.OR)

            # Get the authorized SQL query
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_sql(conn,
                    "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertTrue('references a column' in str(ex_ctx.exception))

    def test_invalid_collection_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Array equality
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "aspects.variables = variables", TEST_ROLE))
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_assets(
                    "list filter", conn, TEST_USER,
                    [ASSET_FULL_RECORD])
            self.assertTrue('are not comparable' in str(ex_ctx.exception))

            # References two collections that don't have the same path, this is not
            # supported now.
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                "GRANT SELECT ON TABLE %s.assets WHERE %s TO ROLE %s" %\
                (TEST_DB, "aspects.variables.name = variables.value", TEST_ROLE))
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._authorize_assets(
                    "independent collection filter", conn, TEST_USER,
                    [ASSET_FULL_RECORD])
            self.assertTrue('Filter references two independent' in str(ex_ctx.exception))

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT name FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name FROM mindsphere_test.assets WHERE " +
                "aspects.variables.name = variables.value")

    def test_column_redaction(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                ("GRANT SELECT(name, description, timezone, `location`)" +
                "ON TABLE %s.assets TO ROLE %s") %
                (TEST_DB, TEST_ROLE))

            expected = {
                "name": "Millenium Falcon",
                "description": "The ship of Han Solo and Chewbacca",
                "location": {
                  "country": "Austria",
                  "region": "Tyrol",
                  "locality": "Innsbruck",
                  "streetAddress": "Industriest 21 A/II",
                  "postalCode": "6020",
                  "longitude": 53.5125546,
                  "latitude": 9.9763411
                },
                "timezone": "Europe/Berlin"
            }

            """
            ------------------------------ column redaction ----------------------------
            Iterations 2000
            Mean 0.8975216150283813 ms
            50%: 0.9009838104248047 ms
            90%: 0.9360313415527344 ms
            95%: 0.9505748748779297 ms
            99%: 1.0018348693847656 ms
            99.5%: 1.0476112365722656 ms
            99.9%: 20.092010498046875 ms
            """
            filter, records, res = self._authorize_assets(
                "column redaction", conn, TEST_USER,
                [ASSET_FULL_RECORD])
            self.assertTrue(filter is None)
            self.assertEqual(records, [expected])
            self.assertTrue(res.filter is None)
            self.assertEqual(len(res.column_access), 34)
            self.assertTrue(res.column_access['name'].accessible)
            # TODO: maintain case
            self.assertFalse(res.column_access['typeid'].accessible)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name, description, `location`.country as location.country, " +
                "`location`.region as location.region, " +
                "`location`.locality as location.locality, " +
                "`location`.streetaddress as location.streetaddress, " +
                "`location`.postalcode as location.postalcode, " +
                "`location`.longitude as location.longitude, " +
                "`location`.latitude as location.latitude, timezone " +
                "FROM mindsphere_test.assets")

    def test_abac_policy(self):
        """ Test both column pruning and filtering """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.delete_attribute('mindsphere_test', 'v')
            conn.create_attribute('mindsphere_test', 'v')
            conn.assign_attribute('mindsphere_test', 'v', TEST_DB, 'assets', 'name')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'description')
            conn.execute_ddl(
                ("GRANT SELECT " +
                "ON TABLE %s.assets HAVING ATTRIBUTE mindsphere_test.v " +
                "WHERE %s " +
                "TO ROLE %s") %
                (TEST_DB, "name = 'Millenium Falcon'", TEST_ROLE))

            expected = {
                "name": "Millenium Falcon",
                "description": "The ship of Han Solo and Chewbacca"
            }

            """
            ---------------------------- abac redaction --------------------------
            Iterations 2000
            Mean 1.3247164487838745 ms
            50%: 1.2581348419189453 ms
            90%: 1.3637542724609375 ms
            95%: 1.7092227935791016 ms
            99%: 1.9984245300292969 ms
            99.5%: 2.070188522338867 ms
            99.9%: 5.983114242553711 ms
            """
            filter, records, res = self._authorize_assets(
                "abac redaction", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(records, [expected])
            self.assertEqual(len(res.filters.filters), 1)
            self.assertEqual(res.filters.filters[0], "name = 'Millenium Falcon'")
            self.assertEqual(res.filters.op, TCompoundOp.OR)
            self.assertEqual(len(res.column_access), 34)
            self.assertTrue(res.column_access['name'].accessible)
            self.assertTrue(res.column_access['description'].accessible)
            self.assertFalse(res.column_access['subtenant'].accessible)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT name, description FROM mindsphere_test.assets " +
                "WHERE name = 'Millenium Falcon'")

    def test_abac_transform(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.delete_attribute('mindsphere_test', 'v')
            conn.create_attribute('mindsphere_test', 'v')
            conn.assign_attribute('mindsphere_test', 'v', TEST_DB, 'assets', 'name')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'description')
            conn.execute_ddl(
                ("GRANT SELECT " +
                "ON TABLE %s.assets HAVING ATTRIBUTE mindsphere_test.v " +
                "TRANSFORM mindsphere_test.v WITH upper() " +
                "TO ROLE %s") %
                (TEST_DB, TEST_ROLE))

            expected = [
                {
                    "name": "MILLENIUM FALCON",
                    "description": "THE SHIP OF HAN SOLO AND CHEWBACCA"
                },
                {
                    "name": "MILLENIUM FALCON2",
                    "description": "THE SHIP OF HAN SOLO AND CHEWBACCA"
                }]

            """
            ---------------------------- abac transform --------------------------
            Iterations 2000
            Mean 1.6276593208312988 ms
            50%: 1.5292167663574219 ms
            90%: 1.8742084503173828 ms
            95%: 1.9330978393554688 ms
            99%: 2.3374557495117188 ms
            99.5%: 2.4857521057128906 ms
            99.9%: 6.550788879394531 ms
            """
            filter, records, res = self._authorize_assets(
                "abac transform", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            self.assertEqual(records, expected)
            self.assertTrue(res.filters is None)
            self.assertFalse(res.column_access['name'].accessible)
            self.assertFalse(res.column_access['description'].accessible)

            # Get the authorized SQL query
            sql = self._authorize_sql(conn,
                "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
            self.assertEqual(sql,
                "SELECT upper(name) as name, upper(description) as description " +
                "FROM mindsphere_test.assets")

    def test_all_transform(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.delete_attribute('mindsphere_test', 'v')
            conn.create_attribute('mindsphere_test', 'v')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'name')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'location.region')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'location.latitude')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'variables.value')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'aspects.variables.name')
            conn.assign_attribute(
                'mindsphere_test', 'v', TEST_DB, 'assets', 'hierarchy.name')
            conn.execute_ddl(
                ("GRANT SELECT " +
                "ON TABLE %s.assets HAVING ATTRIBUTE mindsphere_test.v " +
                "TRANSFORM mindsphere_test.v WITH tokenize() " +
                "TO ROLE %s") %
                (TEST_DB, TEST_ROLE))

            expected = [
                {
                  "aspects": [
                    { "variables": [ { "name": "kjquw" } ] }
                  ],
                  "name": "Mkoykyajx Mzqhfu",
                  "location": {
                    "latitude": 0.2897330143551365,
                    "region": "Jjaei"
                  },
                  "variables": [ { "value": "otcoed" } ]
                },
                {
                  "aspects": [
                    {
                      "variables": [
                        { "name": "kjquw" },
                        { "name": "o" }
                      ]
                    }
                  ],
                  "name": "Kiusuotqo Gyefiu3",
                  "location": {
                    "latitude": 0.2897330143551365,
                    "region": "Jjaei"
                  },
                  "variables": [
                    { "value": "otcoed" }
                  ]
                }
              ]

            """
            ---------------------------- all transform --------------------------
            Iterations 2000
            Mean 2.1513285636901855 ms
            50%: 1.9679069519042969 ms
            90%: 2.796173095703125 ms
            95%: 3.0193328857421875 ms
            99%: 3.673553466796875 ms
            99.5%: 3.9212703704833984 ms
            99.9%: 7.477521896362305 ms
            """
            filter, records, _ = self._authorize_assets(
                "all transform", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            if COMPLEX_TYPES:
              self.assertEqual(records, expected)
              # Get the authorized SQL query
              sql = self._authorize_sql(conn,
                  "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
              self.assertEqual(sql,
                  "SELECT tokenize(name) as name FROM mindsphere_test.assets")

    def test_nested_column_redaction(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                ("GRANT SELECT(name, description, timezone, `location`.country)" +
                "ON TABLE %s.assets TO ROLE %s") %
                (TEST_DB, TEST_ROLE))

            expected1 = {
                "name": "Millenium Falcon",
                "description": "The ship of Han Solo and Chewbacca",
                "location": {
                  "country": "Austria"
                },
                "timezone": "Europe/Berlin"
            }
            expected2 = {
                "name": "Millenium Falcon2",
                "description": "The ship of Han Solo and Chewbacca",
                "location": {
                  "country": "Germany"
                },
                "timezone": "Europe/Berlin"
            }

            """
            ------------------------- nested column redaction ------------------------
            Iterations 2000
            Mean 0.9464281797409058 ms
            50%: 0.9407997131347656 ms
            90%: 0.9677410125732422 ms
            95%: 0.9849071502685547 ms
            99%: 1.0006427764892578 ms
            99.5%: 1.0123252868652344 ms
            99.9%: 5.828619003295898 ms
            """
            filter, records, _ = self._authorize_assets(
                "nested column redaction", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            if COMPLEX_TYPES:
                self.assertEqual(records, [expected1, expected2])

            # Not enabled by default
            #sql = self._authorize_sql(conn,
            #    "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
            #self.assertEqual(sql,
            #    "SELECT upper(name) as name, upper(description) as description " +
            #    "FROM mindsphere_test.assets")

    def test_nested_collection_redaction(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
            conn.execute_ddl(
                ("GRANT SELECT(name, `location`.country, aspects.variables.name)" +
                "ON TABLE %s.assets TO ROLE %s") %
                (TEST_DB, TEST_ROLE))

            expected1 = {
                "name": "Millenium Falcon",
                "location": {
                  "country": "Austria"
                },
                "aspects": [{
                  "variables": [
                    {"name": "color"}
                  ]
                }]
            }
            expected2 = {
                "name": "Millenium Falcon2",
                "location": {
                  "country": "Germany"
                },
                "aspects": [{
                  "variables": [
                    {"name": "color"},
                    {"name": "k"},
                  ]
                }]
            }

            """
            --------------------- nested collection redaction -----------------------
            Iterations 2000
            Mean 0.9543460607528687 ms
            50%: 0.9477138519287109 ms
            90%: 0.9794235229492188 ms
            95%: 0.9908676147460938 ms
            99%: 1.008749008178711 ms
            99.5%: 1.0216236114501953 ms
            99.9%: 7.871389389038086 ms
            """
            filter, records, _ = self._authorize_assets(
                "nested collection redaction", conn, TEST_USER,
                [ASSET_FULL_RECORD, ASSET_RECORD_GERMANY])
            if COMPLEX_TYPES:
                self.assertEqual(records, [expected1, expected2])
                sql = self._authorize_sql(conn,
                    "SELECT * FROM %s.assets" % TEST_DB, user=TEST_USER)
                # FIXME: this is not right
                self.assertEqual(sql,
                    "SELECT name FROM mindsphere_test.assets")

if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import


from okera import odas, _default_context
import boto3
import botocore

_original_make_api_call = None

def patch_botocore():
    global _original_make_api_call
    _original_make_api_call = botocore.client.BaseClient._make_api_call
    botocore.client.BaseClient._make_api_call = _patched_make_api_call

def _patched_make_api_call(client, operation_name, api_params):
    if _default_context and operation_name == 'GetObject':
        # TODO: Determine whether the S3 path is being managed by Okera and
        # only call the Okera if it is
        return _make_get_object_api_call(_default_context, client, api_params)
    return _original_make_api_call(client, operation_name, api_params)

def _make_get_object_api_call(planner, client, api_params):
    path = 's3://{0}/{1}'.format(api_params['Bucket'], api_params['Key'])
    urllib_response = planner.open(path,
                                   version=api_params.get('VersionId', None),
                                   preload_content=False)
    shape = client._service_model.shape_for('GetObjectOutput')
    op_model = client._service_model.operation_model('GetObject')
    http_response = botocore.awsrequest.AWSResponse(path,
                                                    urllib_response.status,
                                                    urllib_response.headers,
                                                    urllib_response)
    response_dict = botocore.endpoint.convert_to_response_dict(http_response, op_model)
    return client._response_parser.parse(response_dict, shape)



# Copyright 2020 Okera Inc. All Rights Reserved.

import boto3
import botocore
import requests

from botocore.credentials import DeferredRefreshableCredentials, CredentialProvider

class OkeraException(Exception):
    pass

class OkeraCredentialProvider(CredentialProvider):
    CANONICAL_NAME = "okera-aws-creds"

    def __init__(self, rest_api_url, token, verify_ssl=False):
        super().__init__()
        self._okera_rest_api_url = rest_api_url.rstrip('/')
        self._okera_token = token
        self._verify_ssl = verify_ssl

    def load(self):
        creds = DeferredRefreshableCredentials(refresh_using=self._refresh, method="sts-assume-role")
        return creds

    def _refresh(self):
        response = self._custom_aws_cred_refresh()
        credentials = {
            "access_key": response.get("key"),
            "secret_key": response.get("secret"),
            "expiry_time": response.get("expiry"),
            "token": None,
        }
        return credentials

    def _custom_aws_cred_refresh(self):
        api_url = "%s/%s" % (self._okera_rest_api_url, "api/v2/aws-tokens")
        headers = {"Authorization": "Bearer %s" % self._okera_token}
        res = requests.post(api_url, headers=headers, verify=self._verify_ssl)
        if res.status_code in (401, 403, 405):
            raise OkeraException("Error in authenticating credentials request: %s" % res.text)
        else:
            return res.json()

def okera_session(session, token, rest_uri, proxy_uri):
    bc_session = None
    if isinstance(session, botocore.session.Session):
        bc_session = session
    else:
        bc_session = session._session

    boto3_session = session

    # TODO: don't assume it's a cred resolver
    cred_chain = bc_session.get_component('credential_provider')
    okera_cred_provider = OkeraCredentialProvider(rest_uri, token)
    if cred_chain.providers:
        first_provider = cred_chain.providers[0].METHOD
        cred_chain.insert_before(first_provider, okera_cred_provider)
    else:
        cred_chain.providers.insert(0, okera_cred_provider)
    config = botocore.config.Config(
        proxies={'https': proxy_uri})

    orig_config = bc_session.get_default_client_config()
    if not orig_config:
        orig_config = botocore.config.Config()

    bc_session.set_default_client_config(orig_config.merge(config))

    return boto3_session


# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Some scenario tests for steward delegation
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

from okera import context
from okera.tests import pycerebro_test_common as common

from okera import _thrift_api

# User and role that are stewards
ROLE = "steward_test_role"
ROLE_MANAGING = "steward_test_role_managing"
USER = "stewardtestuser"
USER_ROLE = "_okera_internal_role_%s" % USER
USER_NO_ACCESS = "stewardtestusernoaccess"
USER_NORMAL = 'stewardtestusernormal'
USER_NORMAL_ROLE = "_okera_internal_role_%s" % USER_NORMAL

# Objects granted stewardship over
CXN = "steward_test_connection"
DB = "steward_test_db"
TBL = "%s.steward_test_table" % (DB)

class StewardTest(common.TestBase):
    # Test granting AS OWNER on the catalog for creating
    # both connections and databases
    def test_steward_1(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (USER_ROLE),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                # Grant the role the right permissions
                "GRANT CREATE_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),
                "GRANT CREATE_DATACONNECTION_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Create the connection, DB and table
            ddls = [
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),

                "CREATE DATABASE %s" % (DB),
                """CREATE EXTERNAL TABLE %s STORED as JDBC
                        TBLPROPERTIES(
                        'driver' = 'mysql',
                        'okera.connection.name' = '%s',
                        'jdbc.db.name'='jdbc_test',
                        'jdbc.schema.name'='public',
                        'table' = 'filter_pushdown_test'
                )""" % (TBL, CXN)
            ]

            # First, do a check that the "no access" user does not
            # have access to create the objects
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)

            for ddl in ddls:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl(ddl)
                self.assertTrue('not have privileges' in str(ex_ctx.exception))

            # Now, check on the real user
            ctx.enable_token_auth(token_str=USER)

            for ddl in ddls:
                conn.execute_ddl(ddl)

            res = conn.scan_as_json('select bigint_col from %s WHERE smallint_col=2' % (TBL))
            assert len(res) == 1
            assert res[0]['bigint_col'] == 4

    # Test granting AS OWNER on the catalog for creating
    # connections, but pre-create the DB and grant access
    # on it (we use CREATE_AS_OWNER so the user can also
    # select on it)
    def test_steward_2(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (USER_ROLE),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                # Create the DB
                "CREATE DATABASE %s" % (DB),

                # Grant the role the right permissions
                "GRANT CREATE_DATACONNECTION_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),
                "GRANT CREATE_AS_OWNER ON DATABASE %s TO ROLE %s" % (DB, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Create the connection, DB and table
            ddls = [
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),

                """CREATE EXTERNAL TABLE %s STORED as JDBC
                        TBLPROPERTIES(
                        'driver' = 'mysql',
                        'okera.connection.name' = '%s',
                        'jdbc.db.name'='jdbc_test',
                        'jdbc.schema.name'='public',
                        'table' = 'filter_pushdown_test'
                )""" % (TBL, CXN)
            ]

            # First, do a check that the "no access" user does not
            # have access to create the objects
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)

            for ddl in ddls:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl(ddl)
                self.assertTrue('not have privileges' in str(ex_ctx.exception))

            # Now, check on the real user
            ctx.enable_token_auth(token_str=USER)

            for ddl in ddls:
                conn.execute_ddl(ddl)

            res = conn.scan_as_json('select bigint_col from %s WHERE smallint_col=2' % (TBL))
            assert len(res) == 1
            assert res[0]['bigint_col'] == 4

    # Test ensuring that a steward can create tables using
    # CREATE TABLE LIKE FILE
    def test_steward_3(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (USER_ROLE),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                # Create the DB
                "CREATE DATABASE %s" % (DB),

                # Grant the role the right permissions
                "GRANT CREATE_AS_OWNER ON DATABASE %s TO ROLE %s" % (DB, ROLE),
                "GRANT ALL ON URI 's3://cerebrodata-test/' TO ROLE %s" % (ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Create the connection, DB and table
            ddls = [
                """CREATE EXTERNAL TABLE %s
                   LIKE PARQUET 's3://cerebrodata-test/alltypes_parquet'
                   STORED AS PARQUET
                   LOCATION 's3://cerebrodata-test/alltypes_parquet'""" % (TBL)
            ]

            # First, do a check that the "no access" user does not
            # have access to create the objects
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)

            for ddl in ddls:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl(ddl)
                self.assertTrue('not have privileges' in str(ex_ctx.exception))

            # Now, check on the real user
            ctx.enable_token_auth(token_str=USER)

            for ddl in ddls:
                conn.execute_ddl(ddl)

            res = conn.scan_as_json('select count(*) as cnt from %s' % (TBL))
            assert len(res) == 1
            assert res[0]['cnt'] == 8

    # End-to-end test including role management delegation:
    # 1. Grant the steward the following abilities:
    #       * Creating roles
    #       * Creating tables in our test DB
    #       * URI access on the necessary URI
    #       * Grant option on the test DB
    # 2. Steward creates a test role, creates a table, and grants
    #    access on the table to the normal user
    # 3. Normal user can now run the query
    def test_steward_4(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (USER_ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE_MANAGING),
                "DROP ROLE IF EXISTS %s" % (USER_NORMAL_ROLE),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                # Create the DB
                "CREATE DATABASE %s" % (DB),

                # Grant the role the right permissions
                "GRANT CREATE_AS_OWNER ON DATABASE %s TO ROLE %s" % (DB, ROLE),
                "GRANT ALL ON URI 's3://cerebrodata-test/' TO ROLE %s" % (ROLE),
                "GRANT CREATE_ROLE_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),
                "GRANT SELECT ON DATABASE %s TO ROLE %s WITH GRANT OPTION" % (DB, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Create the connection, DB and table
            ddls = [
                "CREATE ROLE %s" % (ROLE_MANAGING),

                """CREATE EXTERNAL TABLE %s
                   LIKE PARQUET 's3://cerebrodata-test/alltypes_parquet'
                   STORED AS PARQUET
                   LOCATION 's3://cerebrodata-test/alltypes_parquet'""" % (TBL),

                "GRANT SELECT ON TABLE %s TO ROLE %s" % (TBL, ROLE_MANAGING),
                "GRANT ROLE %s TO GROUP %s" % (ROLE_MANAGING, USER_NORMAL),
            ]

            # First, do a check that the "no access" user does not
            # have access to create the objects
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)

            for ddl in ddls:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl(ddl)
                self.assertTrue(
                    'not have privileges' in str(ex_ctx.exception) or
                    'does not exist' in str(ex_ctx.exception))

            # Now, check on the real user
            ctx.enable_token_auth(token_str=USER)

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Now, switch to a normal user (who was just granted access to this
            # table), and verify they can run a query
            ctx.enable_token_auth(token_str=USER_NORMAL)

            res = conn.scan_as_json('select count(*) as cnt from %s' % (TBL))
            assert len(res) == 1
            assert res[0]['cnt'] == 8

            # Verify the no access user can't run the query
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)

            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.scan_as_json('select count(*) as cnt from %s' % (TBL))
            self.assertTrue('not have privileges' in str(ex_ctx.exception))


# Copyright 2017 Okera Inc. All Rights Reserved.
#
# Tests that should run on any configuration. The server auth can be specified
# as an environment variables before running this test.

# pylint: disable=no-member
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods
# pylint: disable=bad-continuation
# pylint: disable=bad-indentation
import time
import unittest
import string
import random

from okera.tests import pycerebro_test_common as common

TEST_DB = 'access_review_test_db'

def generate_random_users(num):
    users = []
    for _ in range(num):
        res = ''.join(random.choice(string.ascii_lowercase)  for i in range(5))
        user = 'user_' + res
        users.append(user)
    return users

def generate_random_tables(conn, num):
    tables = []
    for _ in range(num):
        res = ''.join(random.choice(string.ascii_lowercase)  for i in range(5))
        table_name = '%s.table_%s' % (TEST_DB, res)
        tables.append(table_name)
        conn.execute_ddl('CREATE TABLE %s (col1 int)' % table_name)
    return tables

def get_users_and_groups(ctx, conn, expected_users=None, tries=5):
    attempt = 0
    expected_users = expected_users or []
    while attempt < tries:
        ctx.enable_token_auth(token_str='root')
        results = conn.scan_as_json('select * from okera_system.groups_by_user')
        users_to_groups = {}
        for result in results:
            groups = users_to_groups.get(result['user_name'], set())
            groups.add(result['group_name'])
            users_to_groups[result['user_name']] = groups

        found_users = 0
        for expected_user in expected_users:
            if expected_user in users_to_groups:
                found_users += 1
        if found_users == len(expected_users):
            return users_to_groups

        time.sleep(2)
        attempt += 1

    raise Exception("Couldn't find users '%s' after %s tries" % (expected_users, tries))

def get_access(ctx, conn, expected_accesses=None, tries=5):
    attempt = 0
    expected_accesses = expected_accesses or []
    while attempt < tries:
        ctx.enable_token_auth(token_str='root')
        results = conn.scan_as_json(
            'select * from okera_system.catalog_identity_mappings')
        access_mappings = {}
        for result in results:
            objects = access_mappings.get(result['identity_object_name'], set())
            objects.add(result['catalog_object_name'])
            access_mappings[result['identity_object_name']] = objects

        found_accesses = 0
        for expected_access in expected_accesses:
            if expected_access['user'] in access_mappings:
                found_tables = 0
                for obj in expected_access['objects']:
                    if obj not in access_mappings[expected_access['user']]:
                        continue
                    found_tables += 1

                if found_tables == len(expected_access['objects']):
                    found_accesses += 1

        if found_accesses == len(expected_accesses):
            return access_mappings

        time.sleep(2)
        attempt += 1

    raise Exception(
        "Couldn't find mappings '%s' after %s tries" % (expected_accesses, tries))


class AccessReviewTest(unittest.TestCase):

    def test_basic_access_review(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % TEST_DB)
            conn.execute_ddl('CREATE DATABASE %s' % TEST_DB)

            random_users = generate_random_users(3)
            random_tables = generate_random_tables(conn, 3)
            random_accesses = [dict(user=u, objects=random_tables) for u in random_users]

            for access in random_accesses:
                role_name = 'access_review_test_role_%s' % access['user']
                conn.execute_ddl('CREATE ROLE IF NOT EXISTS %s' % role_name)
                conn.execute_ddl(
                    'GRANT ROLE %s TO GROUP %s' % (role_name, access['user']))
                for table in access['objects']:
                    conn.execute_ddl(
                        'GRANT SELECT ON TABLE %s TO ROLE %s' % (table, role_name))

            for user in random_users:
                ctx.enable_token_auth(token_str=user)
                for table in random_tables:
                    conn.scan_as_json('select * from %s' % table)

            users_and_groups = get_users_and_groups(
                ctx, conn, expected_users=random_users, tries=16)
            self.assertTrue(users_and_groups is not None)
            accesses = get_access(
                ctx, conn, expected_accesses=random_accesses, tries=16)
            self.assertTrue(accesses is not None)

if __name__ == "__main__":
    unittest.main()



# Copyright 2020 Okera Inc. All Rights Reserved.
import os

from boto3 import Session
from pyspark.sql import SparkSession

OKERA_HOME = os.environ["OKERA_HOME"]
OKERA_JAVA_VERSION = os.environ["OKERA_JAVA_VERSION"]

# Planner vars
ODAS_TEST_HOST = os.environ["ODAS_TEST_HOST"]
PLANNER_THRIFT_URL = ODAS_TEST_HOST + ":" + os.environ['ODAS_TEST_PORT_PLANNER_THRIFT']

# Hive thrift URL
HIVE_THRIFT_URL = 'thrift://' + ODAS_TEST_HOST + ":" + \
    os.environ['ODAS_TEST_PORT_HIVE_THRIFT']
# Spark master URL
SPARK_MASTER_URL = 'spark://' + ODAS_TEST_HOST + ":" + \
    os.environ['ODAS_TEST_PORT_SPARK_MASTER']

# This is needed for authenticating to ODAS planner.
DEFAULT_SPARK_USER = 'okera'
os.environ["HADOOP_USER_NAME"] = DEFAULT_SPARK_USER

# This is needed for HMS discovery. Note, the spark job is still submitted to the
# spark cluster.
SPARK_JAR_NAME = OKERA_HOME + '/rsc/java/spark2/target/recordservice-spark-2.0-' + \
    OKERA_JAVA_VERSION + '.jar'
HIVE_JAR_NAME = OKERA_HOME + '/rsc/java/hive/target/recordservice-hive-' + \
    OKERA_JAVA_VERSION + '.jar'
HIVE_UDF_NAME = OKERA_HOME + '/tools/hive-udf/target/okera-hive-udfs-0.1.1.jar'

# Additional jars needed to read from S3, this is not included by spark by default.
ADDITIONAL_JARS = OKERA_HOME + '/thirdparty/spark/hadoop-aws-2.7.3.jar'
ADDITIONAL_JARS += ',' + OKERA_HOME + '/thirdparty/spark/aws-java-sdk-1.7.4.jar'
ADDITIONAL_JARS += ',' + OKERA_HOME + '/thirdparty/spark/snappy-java-1.1.7.3.jar'

os.environ["PYSPARK_SUBMIT_ARGS"] = '--jars ' + SPARK_JAR_NAME + ',' + HIVE_JAR_NAME + \
    ',' + HIVE_UDF_NAME + ',' + ADDITIONAL_JARS + ' pyspark-shell'

def get_spark_session(app_name):
    credentials = Session().get_credentials()

    return (SparkSession.builder.appName(app_name)
            .master(SPARK_MASTER_URL)
            .config("spark.recordservice.user", DEFAULT_SPARK_USER)
            .config("hive.metastore.uris", HIVE_THRIFT_URL)
            .enableHiveSupport()
            .config("spark.recordservice.planner.hostports",
                    PLANNER_THRIFT_URL)
            .config("fs.s3a.access.key", credentials.access_key)
            .config("fs.s3a.secret.key", credentials.secret_key)
            .config("spark.cores.max", "2")
            .config("spark.executor.memory", "2048M")
            .getOrCreate())

authorize_query_unsupported_datasets = [
    # Complex types view syntax issues
    'authdb.struct_t_join_clause_view',
    'chase.feescharges_view',
    'chase.second_level',
    'chase.subscription_party_view',
    'rs_complex.pn_view',
    'zd1179large.store_sales_pk_view',

    # Unknown, need to explore
    'demo_test.transactions_active_users',
    'demo_test.transactions_anonymize_inactive_users',
    'nytaxi.unionallview',
    'rs_complex.market_decide_single_avro',

    # Join hint is causing issues, should not be output in query rewrite
    'rs.nation_join_cache',
    'rs.nation_join_cache_and_size_hint',
    'rs.nation_join_size_hint',

    # Skipping as these joins take a long time
    'tpch_sf1.lineitem_orders',
    'tpch_sf1.lineitem_orders_medium',
    'tpch_sf5.lineitem_orders',
    'tpch_sf5.lineitem_orders_medium',
    'tpch_sf5.lineitem_orders_small',

    # Hive/Spark doesn't like the avro schemas
    'chase.zd1211',
    'chase.zd1211_1',
    'chase.zd1211_join_view',
    'chase.zd1211_view',
    'chase.zd1238',
    'chase.zd1238_1',
    'chase.zd1238_2',
    'chase.zd1238_3',
    'chase.zd1238_4',
    'chase.zd1238_5',
    'chase.zd1238_6',
    'chase.zd1238_7',
    'chase.zd1238_8',
    'chase.zd1238_9',
    'customer.account_alternate1_phone_number_updated_from_file',
    'rs_complex.avrotbl',
    'rs_complex.bytes_type_file',
    'rs_complex.global_dtc_daily_store_stock_on_hand_snapshot',
    'rs_complex.non_null_union',
    'rs_complex.rs_complex_array_map_t',
    'rs_complex.zd1216',
    'rs_complex.zd1216_with_subscriptionlimit',

    # TODO: complex views
    'rs_complex.struct_nested_s1_s2',
    'rs_complex_parquet.struct_nested_s1_s2',
]
